# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'liquidity_rebalancing_abci' skill."""
import json
from abc import ABC
from typing import Any, Dict, Generator, List, Optional, Set, Type, cast

from hexbytes import HexBytes

from packages.valory.contracts.gnosis_safe.contract import (
    GnosisSafeContract,
    SafeOperation,
)
from packages.valory.contracts.multisend.contract import (
    MultiSendContract,
    MultiSendOperation,
)
from packages.valory.contracts.uniswap_v2_erc20.contract import UniswapV2ERC20Contract
from packages.valory.contracts.uniswap_v2_router_02.contract import (
    UniswapV2Router02Contract,
)
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.liquidity_rebalancing_abci.models import Params, SharedState
from packages.valory.skills.liquidity_rebalancing_abci.payloads import (
    SleepPayload,
    StrategyEvaluationPayload,
    StrategyType,
    TransactionHashPayload,
)
from packages.valory.skills.liquidity_rebalancing_abci.rounds import (
    EnterPoolTransactionHashRound,
    ExitPoolTransactionHashRound,
    LiquidityRebalancingAbciApp,
    SleepRound,
    StrategyEvaluationRound,
    SwapBackTransactionHashRound,
    SynchronizedData,
)
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    hash_payload_to_hex,
)


# These safeTxGas values are calculated from experimental values plus
# a 10% buffer and rounded up. The Gnosis safe default value is 0 (max gas)
# https://help.gnosis-safe.io/en/articles/4738445-advanced-transaction-parameters
# More on gas estimation: https://help.gnosis-safe.io/en/articles/4933491-gas-estimation
SAFE_TX_GAS_ENTER = 553000
SAFE_TX_GAS_EXIT = 248000
SAFE_TX_GAS_SWAP_BACK = 268000


def parse_tx_token_balance(
    transfer_logs: List[Dict],
    token_address: str,
    source_address: str,
    destination_address: str,
) -> int:
    """
    Returns the transfered token amount from one address to another given a list of transactions.

    :param transfer_logs: a list of transactions.
    :param token_address: the token address.
    :param source_address: the source address.
    :param destination_address: the destination address.
    :return: the transfered amount
    """

    token_events = list(
        filter(
            lambda log: log["token_address"] == token_address
            and log["from"] == source_address
            and log["to"] == destination_address,
            transfer_logs,
        )
    )
    return sum(event["value"] for event in token_events)


class LiquidityRebalancingBaseBehaviour(BaseBehaviour, ABC):
    """Base behaviour for the liquidity rebalancing skill."""

    @property
    def synchronized_data(self) -> SynchronizedData:
        """Return the synchronized data."""
        return cast(SynchronizedData, super().synchronized_data)

    @property
    def params(self) -> Params:
        """Return the params."""
        return cast(Params, self.context.params)

    def get_swap_tx_data(  # pylint: disable=too-many-arguments
        self,
        is_input_native: bool,
        is_output_native: bool,
        exact_input: bool,
        path: List[str],
        deadline: int,
        amount_in: Optional[int] = None,
        amount_out: Optional[int] = None,
        amount_in_max: Optional[int] = None,
        amount_out_min: Optional[int] = None,
        eth_value: Optional[int] = None,
    ) -> Generator[None, None, Optional[Dict]]:
        """
        Return the swap tx data.

        :param is_input_native: flag to check if the first token is native
        :param is_output_native: flag to check if the second token is native
        :param exact_input: flag to check if the exact amount belongs to the input or the output
        :param path: the swap path
        :param deadline: the tx deadline
        :param amount_in: the amount in
        :param amount_out: the amount out
        :param amount_in_max: the max amount in
        :param amount_out_min: the min amount out
        :param eth_value: the tx value, also used as amount_or amount_in_max in for some cases
        :return: None if required parameters are missing
        :yield: the tx data
        """

        if (  # pylint: disable=too-many-boolean-expressions
            (is_input_native and is_output_native)
            or (is_input_native and eth_value is None)
            or (not is_input_native and exact_input and amount_in is None)
            or (not is_input_native and not exact_input and amount_in_max is None)
            or (exact_input and amount_out_min is None)
            or (not exact_input and amount_out is None)
        ):
            self.context.logger.error("Swap data is not correct.")
            raise RuntimeError("Swap has been called with incorrect/missing parameters")
        method_name = (
            f'swap_exact_{"ETH" if is_input_native else "tokens"}_for_{"ETH" if is_output_native else "tokens"}'
            if exact_input
            else f'swap_{"ETH" if is_input_native else "tokens"}_for_exact_{"ETH" if is_output_native else "tokens"}'
        )

        contract_api_kwargs: Dict[str, Any] = dict(
            method_name=method_name,
            path=path,
            to=self.synchronized_data.safe_contract_address,
            deadline=deadline,
        )

        # Input amounts for native tokens are read from the msg.value field.
        # We only need to specify them here for not native tokens.
        if not is_input_native and exact_input:
            contract_api_kwargs["amount_in"] = int(amount_in)  # type: ignore

        if not is_input_native and not exact_input:
            contract_api_kwargs["amount_in_max"] = int(amount_in_max)  # type: ignore

        # Output amounts
        if exact_input:
            contract_api_kwargs["amount_out_min"] = int(amount_out_min)  # type: ignore
        else:
            contract_api_kwargs["amount_out"] = int(amount_out)  # type: ignore

        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.synchronized_data.router_contract_address,
            contract_id=str(UniswapV2Router02Contract.contract_id),
            contract_callable="get_method_data",
            **contract_api_kwargs,
        )
        swap_data = cast(bytes, contract_api_msg.raw_transaction.body["data"])

        return {
            "operation": MultiSendOperation.CALL,
            "to": self.synchronized_data.router_contract_address,
            "value": eth_value
            if is_input_native
            else 0,  # Input amount for native tokens
            "data": HexBytes(swap_data.hex()),
        }

    def get_swap_data(
        self, strategy: dict, token: str, is_swap_back: bool
    ) -> Generator[None, None, Optional[Dict]]:
        """
        Return the swap tx data for swaps, particularized for swaps base->token and token->base.

        :param strategy: the strategy
        :param token: "token_a" or "token_b"
        :param is_swap_back: True for token[a,b] -> token_base, False for token_base -> token[a,b]
        :return: the tx data
        """

        token_letter = token[-1]  # a or b
        input_token = token if is_swap_back else "token_base"
        output_token = "token_base" if is_swap_back else token

        kwargs = dict(
            deadline=strategy["deadline"],
            is_input_native=strategy[input_token]["is_native"],
            is_output_native=strategy[output_token]["is_native"],
            exact_input=is_swap_back,
            path=[strategy[t]["address"] for t in (input_token, output_token)],
        )

        if not is_swap_back:
            kwargs["amount_out"] = strategy[token]["amount_after_swap"]
            kwargs["amount_in_max"] = strategy["token_base"][
                f"amount_in_max_{token_letter}"
            ]
            kwargs["eth_value"] = (
                strategy["token_base"][f"amount_in_max_{token_letter}"]
                if strategy["token_base"]["is_native"]
                else 0
            )

        else:
            kwargs["amount_in"] = strategy[token]["amount_received"]
            kwargs["amount_out_min"] = strategy["token_base"][
                f"amount_min_after_swap_back_{token_letter}"
            ]
            kwargs["eth_value"] = (
                strategy[token]["amount_received"]
                if strategy[token]["is_native"]
                else 0
            )

        return self.get_swap_tx_data(**kwargs)

    def get_tx_result(self) -> Generator[None, None, list]:
        """Transaction transfer result."""
        strategy = json.loads(self.synchronized_data.most_voted_strategy)
        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_address=strategy["token_LP"]["address"],
            contract_id=str(UniswapV2ERC20Contract.contract_id),
            contract_callable="get_transaction_transfer_logs",
            tx_hash=self.synchronized_data.final_tx_hash,
            target_address=self.synchronized_data.safe_contract_address,
        )
        if contract_api_msg.performative != ContractApiMessage.Performative.STATE:
            self.context.logger.info(
                f"Error retrieving the transaction logs for hash: {self.synchronized_data.final_tx_hash}"
            )
            return []  # pragma: nocover
        transfers = cast(list, contract_api_msg.state.body["logs"])

        transfer_log_message = (
            f"The tx with hash {self.synchronized_data.final_tx_hash} ended with the following transfers.\n"
            f"Transfers: {str(transfers)}\n"
        )
        self.context.logger.info(transfer_log_message)
        return transfers

    def get_allowance_data(
        self, token_address: str, value: int
    ) -> Generator[None, None, dict]:
        """
        Return the swap tx data for swaps, particularized for swaps base->token and token->base.

        :param token_address: the spender's address
        :param value: the allowance value to set
        :return: the tx data
        :yield: the tx data
        """

        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=token_address,
            contract_id=str(UniswapV2ERC20Contract.contract_id),
            contract_callable="get_method_data",
            method_name="approve",
            spender=self.synchronized_data.router_contract_address,
            value=value,
        )

        return {
            "operation": MultiSendOperation.CALL,
            "to": token_address,
            "value": 0,
            "data": HexBytes(
                cast(bytes, contract_api_msg.raw_transaction.body["data"]).hex()
            ),
        }


class StrategyEvaluationBehaviour(LiquidityRebalancingBaseBehaviour):
    """Evaluate the financial strategy."""

    behaviour_id = "strategy_evaluation"
    matching_round = StrategyEvaluationRound

    def async_act(self) -> Generator:
        """Do the action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():

            # Get the previous strategy or use the dummy one
            # For now, the app will loop between enter-exit-swap_back,
            # unless we start with WAIT. Then it will keep waiting.
            strategy: dict = {}
            try:
                strategy = json.loads(self.synchronized_data.most_voted_strategy)

                if strategy["action"] == StrategyType.ENTER.value:
                    strategy["action"] = StrategyType.EXIT.value

                elif strategy["action"] == StrategyType.EXIT.value:
                    strategy["action"] = StrategyType.SWAP_BACK.value

                elif strategy["action"] == StrategyType.SWAP_BACK.value:
                    strategy["action"] = StrategyType.ENTER.value

            # An exception will occur during the first run as no strategy was set
            except ValueError:
                strategy = self.get_dummy_strategy()

            # Log the new strategy
            if strategy["action"] == StrategyType.WAIT.value:  # pragma: nocover
                self.context.logger.info("Current strategy is still optimal. Waiting.")

            if strategy["action"] == StrategyType.ENTER.value:
                self.context.logger.info(
                    "Performing strategy update: moving into "
                    + f"{strategy['token_a']['ticker']}-{strategy['token_b']['ticker']} (pool {self.synchronized_data.router_contract_address})"
                )

            if strategy["action"] == StrategyType.EXIT.value:  # pragma: nocover
                self.context.logger.info(
                    "Performing strategy update: moving out of "
                    + f"{strategy['token_a']['ticker']}-{strategy['token_b']['ticker']} (pool {self.synchronized_data.router_contract_address})"
                )

            if strategy["action"] == StrategyType.SWAP_BACK.value:  # pragma: nocover
                self.context.logger.info(
                    f"Performing strategy update: swapping back {strategy['token_a']['ticker']}, {strategy['token_b']['ticker']}"
                )

            payload = StrategyEvaluationPayload(
                self.context.agent_address, json.dumps(strategy, sort_keys=True)
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()

    def get_dummy_strategy(self) -> dict:
        """Get a dummy strategy."""
        last_timestamp = cast(
            SharedState, self.context.state
        ).round_sequence.abci_app.last_timestamp.timestamp()

        strategy = {
            "action": StrategyType.ENTER.value,
            "safe_nonce": 0,
            "safe_tx_gas": {
                "enter": SAFE_TX_GAS_ENTER,
                "exit": SAFE_TX_GAS_EXIT,
                "swap_back": SAFE_TX_GAS_SWAP_BACK,
            },
            "deadline": int(last_timestamp)
            + self.params.rebalancing_params["deadline"],
            "chain": self.params.rebalancing_params["chain"],
            "token_base": {
                "ticker": self.params.rebalancing_params["token_base_ticker"],
                "address": self.params.rebalancing_params["token_base_address"],
                "amount_in_max_a": int(1e4),
                "amount_min_after_swap_back_a": int(1e2),
                "amount_in_max_b": int(1e4),
                "amount_min_after_swap_back_b": int(1e2),
                "is_native": False,
                "set_allowance": self.params.rebalancing_params["max_allowance"],
                "remove_allowance": 0,
            },
            "token_LP": {
                "address": self.params.rebalancing_params["lp_token_address"],
                "set_allowance": self.params.rebalancing_params["max_allowance"],
                "remove_allowance": 0,
            },
            "token_a": {
                "ticker": self.params.rebalancing_params["token_a_ticker"],
                "address": self.params.rebalancing_params["token_a_address"],
                "amount_after_swap": int(1e3),
                "amount_min_after_add_liq": int(0.5e3),
                "is_native": False,  # if one of the two tokens is native, A must be the one
                "set_allowance": self.params.rebalancing_params["max_allowance"],
                "remove_allowance": 0,
                "amount_received_after_exit": 0,
            },
            "token_b": {
                "ticker": self.params.rebalancing_params["token_b_ticker"],
                "address": self.params.rebalancing_params["token_b_address"],
                "amount_after_swap": int(1e3),
                "amount_min_after_add_liq": int(0.5e3),
                "is_native": False,  # if one of the two tokens is native, A must be the one
                "set_allowance": self.params.rebalancing_params["max_allowance"],
                "remove_allowance": 0,
                "amount_received_after_exit": 0,
            },
        }
        return strategy


class SleepBehaviour(LiquidityRebalancingBaseBehaviour):
    """Wait for a predefined amount of time."""

    behaviour_id = "sleep"
    matching_round = SleepRound

    def async_act(self) -> Generator:
        """Do the action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():

            yield from self.sleep(self.params.rebalancing_params["sleep_seconds"])
            payload = SleepPayload(self.context.agent_address)

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class EnterPoolTransactionHashBehaviour(LiquidityRebalancingBaseBehaviour):
    """Prepare the transaction hash for entering the liquidity pool

    The expected transfers derived from this behaviour are
    Safe         ->  A-Base-pool : Base tokens
    A-Base-pool  ->  Safe        : A tokens
    Safe         ->  B-Base-pool : Base tokens
    B-Base-pool  ->  Safe        : B tokens
    Safe         ->  A-B-pool    : A tokens
    Safe         ->  A-B-pool    : B tokens
    A-B-pool Minter       ->  Safe        : AB_LP tokens
    """

    behaviour_id = "enter_pool_tx_hash"
    matching_round = EnterPoolTransactionHashRound

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
        - Request the transaction hash for the safe transaction. This is the
          hash that needs to be signed by a threshold of agents.
        - Send the transaction hash as a transaction and wait for it to be mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():

            strategy = json.loads(self.synchronized_data.most_voted_strategy)

            # Prepare a uniswap tx list. We should check what token balances we have at this point.
            # It is possible that we don't need to swap. For now let's assume we have just USDT
            # and always swap back to it.
            multi_send_txs = []

            # Add allowance for base token
            if (
                not strategy["token_base"]["is_native"]
                and "set_allowance" in strategy["token_base"]
            ):
                allowance_base_data = yield from self.get_allowance_data(
                    token_address=strategy["token_base"]["address"],
                    value=strategy["token_base"]["set_allowance"],
                )

                multi_send_txs.append(allowance_base_data)

            # Swap first token
            if strategy["token_a"]["address"] != strategy["token_base"]["address"]:

                swap_tx_data = yield from self.get_swap_data(  # nosec
                    strategy=strategy, token="token_a", is_swap_back=False
                )
                if swap_tx_data:
                    multi_send_txs.append(swap_tx_data)

            # Swap second token
            if strategy["token_b"]["address"] != strategy["token_base"]["address"]:

                swap_tx_data = yield from self.get_swap_data(  # nosec
                    strategy=strategy, token="token_b", is_swap_back=False
                )
                if swap_tx_data:
                    multi_send_txs.append(swap_tx_data)

            # Add allowance for token A (only if not native)
            if (
                not strategy["token_a"]["is_native"]
                and "set_allowance" in strategy["token_a"]
            ):
                allowance_a_data = yield from self.get_allowance_data(
                    token_address=strategy["token_a"]["address"],
                    value=strategy["token_a"]["set_allowance"],
                )

                multi_send_txs.append(allowance_a_data)

            # Add allowance for token B (only if not native)
            if (
                not strategy["token_b"]["is_native"]
                and "set_allowance" in strategy["token_b"]
            ):
                allowance_b_data = yield from self.get_allowance_data(
                    token_address=strategy["token_b"]["address"],
                    value=strategy["token_b"]["set_allowance"],
                )

                multi_send_txs.append(allowance_b_data)

            # Add liquidity
            if strategy["token_a"]["is_native"]:
                contract_api_msg = yield from self.get_contract_api_response(
                    performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                    contract_address=self.synchronized_data.router_contract_address,
                    contract_id=str(UniswapV2Router02Contract.contract_id),
                    contract_callable="get_method_data",
                    method_name="add_liquidity_ETH",
                    token=strategy["token_b"]["address"],
                    amount_token_desired=int(strategy["token_b"]["amount_after_swap"]),
                    amount_token_min=int(
                        strategy["token_b"]["amount_min_after_add_liq"]
                    ),
                    amount_ETH_min=int(strategy["token_a"]["amount_min_after_add_liq"]),
                    to=self.synchronized_data.safe_contract_address,
                    deadline=strategy["deadline"],
                )
                liquidity_data = cast(
                    bytes, contract_api_msg.raw_transaction.body["data"]
                )
                multi_send_txs.append(
                    {
                        "operation": MultiSendOperation.CALL,
                        "to": self.synchronized_data.router_contract_address,
                        "value": int(strategy["token_a"]["amount_min_after_add_liq"]),
                        "data": HexBytes(liquidity_data.hex()),
                    }
                )

            else:
                contract_api_msg = yield from self.get_contract_api_response(
                    performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                    contract_address=self.synchronized_data.router_contract_address,
                    contract_id=str(UniswapV2Router02Contract.contract_id),
                    contract_callable="get_method_data",
                    method_name="add_liquidity",
                    token_a=strategy["token_a"]["address"],
                    token_b=strategy["token_b"]["address"],
                    amount_a_desired=int(strategy["token_a"]["amount_after_swap"]),
                    amount_b_desired=int(strategy["token_b"]["amount_after_swap"]),
                    amount_a_min=int(strategy["token_a"]["amount_min_after_add_liq"]),
                    amount_b_min=int(strategy["token_b"]["amount_min_after_add_liq"]),
                    to=self.synchronized_data.safe_contract_address,
                    deadline=strategy["deadline"],  # 5 min into the future
                )
                liquidity_data = cast(
                    bytes, contract_api_msg.raw_transaction.body["data"]
                )
                multi_send_txs.append(
                    {
                        "operation": MultiSendOperation.CALL,
                        "to": self.synchronized_data.router_contract_address,
                        "value": 0,
                        "data": HexBytes(liquidity_data.hex()),
                    }
                )

            # Get the tx list data from multisend contract
            contract_api_msg = yield from self.get_contract_api_response(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                contract_address=self.synchronized_data.safe_contract_address,
                contract_id=str(MultiSendContract.contract_id),
                contract_callable="get_tx_data",
                multi_send_txs=multi_send_txs,
            )
            multisend_data = cast(str, contract_api_msg.raw_transaction.body["data"])
            multisend_data = multisend_data[2:]
            self.context.logger.info(f"Multisend data: {multisend_data}")
            # Get the tx hash from Gnosis Safe contract
            contract_api_msg = yield from self.get_contract_api_response(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                contract_address=self.synchronized_data.safe_contract_address,
                contract_id=str(GnosisSafeContract.contract_id),
                contract_callable="get_raw_safe_transaction_hash",
                to_address=self.synchronized_data.multisend_contract_address,
                value=0,
                data=bytes.fromhex(multisend_data),
                operation=SafeOperation.DELEGATE_CALL.value,
                safe_tx_gas=strategy["safe_tx_gas"]["enter"],
                safe_nonce=strategy["safe_nonce"],
            )
            safe_tx_hash = cast(str, contract_api_msg.raw_transaction.body["tx_hash"])
            safe_tx_hash = safe_tx_hash[2:]
            self.context.logger.info(f"Hash of the Safe transaction: {safe_tx_hash}")

            payload_string = hash_payload_to_hex(
                safe_tx_hash=safe_tx_hash,
                ether_value=0,
                safe_tx_gas=strategy["safe_tx_gas"]["enter"],
                to_address=self.synchronized_data.multisend_contract_address,
                data=bytes.fromhex(multisend_data),
                operation=SafeOperation.DELEGATE_CALL.value,
            )

            payload = TransactionHashPayload(
                sender=self.context.agent_address, tx_hash=payload_string
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class ExitPoolTransactionHashBehaviour(LiquidityRebalancingBaseBehaviour):
    """Prepare the transaction hash for exiting the liquidity pool

    The expected transfers derived from this behaviour are
    Safe         ->  A-B-pool    : AB_LP tokens
    AB_LP        ->  Safe        : A tokens
    AB_LP        ->  Safe        : B tokens
    """

    behaviour_id = "exit_pool_tx_hash"
    matching_round = ExitPoolTransactionHashRound

    def async_act(self) -> Generator:  # pylint: disable=too-many-statements
        """
        Do the action.

        Steps:
        - Request the transaction hash for the safe transaction. This is the
          hash that needs to be signed by a threshold of agents.
        - Send the transaction hash as a transaction and wait for it to be mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():

            strategy = json.loads(self.synchronized_data.most_voted_strategy)

            # Get previous transaction's results
            transfers = yield from self.get_tx_result()

            amount_a_sent: int = parse_tx_token_balance(
                transfer_logs=transfers,
                token_address=strategy["token_a"]["address"],
                source_address=self.synchronized_data.safe_contract_address,
                destination_address=self.synchronized_data.router_contract_address,
            )
            amount_b_sent: int = parse_tx_token_balance(
                transfer_logs=transfers,
                token_address=strategy["token_b"]["address"],
                source_address=self.synchronized_data.safe_contract_address,
                destination_address=self.synchronized_data.router_contract_address,
            )
            amount_liquidity_received: int = parse_tx_token_balance(
                transfer_logs=transfers,
                token_address=strategy["token_LP"]["address"],
                source_address=self.params.rebalancing_params["default_minter"],
                destination_address=self.synchronized_data.safe_contract_address,
            )

            # Prepare a uniswap tx list. We should check what token balances we have at this point.
            # It is possible that we don't need to swap. For now let's assume we have just USDT
            # and always swap back to it.
            multi_send_txs = []

            # Add allowance for LP token to be spent by the router
            allowance_lp_data = yield from self.get_allowance_data(
                token_address=strategy["token_LP"]["address"],
                value=strategy["token_LP"]["set_allowance"],
            )

            multi_send_txs.append(allowance_lp_data)

            # Remove liquidity
            if strategy["token_a"]["is_native"]:

                contract_api_msg = yield from self.get_contract_api_response(
                    performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                    contract_address=self.synchronized_data.router_contract_address,
                    contract_id=str(UniswapV2Router02Contract.contract_id),
                    contract_callable="get_method_data",
                    method_name="remove_liquidity_ETH",
                    token=strategy["token_b"]["address"],
                    liquidity=amount_liquidity_received,
                    amount_token_min=int(amount_b_sent),
                    amount_ETH_min=int(amount_a_sent),
                    to=self.synchronized_data.safe_contract_address,
                    deadline=strategy["deadline"],
                )
                liquidity_data = cast(
                    bytes, contract_api_msg.raw_transaction.body["data"]
                )

            else:

                contract_api_msg = yield from self.get_contract_api_response(
                    performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                    contract_address=self.synchronized_data.router_contract_address,
                    contract_id=str(UniswapV2Router02Contract.contract_id),
                    contract_callable="get_method_data",
                    method_name="remove_liquidity",
                    token_a=strategy["token_a"]["address"],
                    token_b=strategy["token_b"]["address"],
                    liquidity=amount_liquidity_received,
                    amount_a_min=int(amount_a_sent),
                    amount_b_min=int(amount_b_sent),
                    to=self.synchronized_data.safe_contract_address,
                    deadline=strategy["deadline"],
                )
                liquidity_data = cast(
                    bytes, contract_api_msg.raw_transaction.body["data"]
                )

            multi_send_txs.append(
                {
                    "operation": MultiSendOperation.CALL,
                    "to": self.synchronized_data.router_contract_address,
                    "value": 0,
                    "data": HexBytes(liquidity_data.hex()),
                }
            )

            # Remove allowance for LP token
            if "remove_allowance" in strategy["token_LP"]:
                allowance_lp_data = yield from self.get_allowance_data(
                    token_address=strategy["token_LP"]["address"],
                    value=strategy["token_LP"]["remove_allowance"],
                )

                multi_send_txs.append(allowance_lp_data)

            # Get the tx list data from multisend contract
            contract_api_msg = yield from self.get_contract_api_response(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                contract_address=self.synchronized_data.safe_contract_address,
                contract_id=str(MultiSendContract.contract_id),
                contract_callable="get_tx_data",
                multi_send_txs=multi_send_txs,
            )
            multisend_data = cast(str, contract_api_msg.raw_transaction.body["data"])
            multisend_data = multisend_data[2:]
            self.context.logger.info(f"Multisend data: {multisend_data}")
            # Get the tx hash from Gnosis Safe contract
            contract_api_msg = yield from self.get_contract_api_response(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                contract_address=self.synchronized_data.safe_contract_address,
                contract_id=str(GnosisSafeContract.contract_id),
                contract_callable="get_raw_safe_transaction_hash",
                to_address=self.synchronized_data.multisend_contract_address,
                value=0,
                data=bytes.fromhex(multisend_data),
                operation=SafeOperation.DELEGATE_CALL.value,
                safe_tx_gas=strategy["safe_tx_gas"]["exit"],
                safe_nonce=strategy["safe_nonce"],
            )
            safe_tx_hash = cast(str, contract_api_msg.raw_transaction.body["tx_hash"])
            safe_tx_hash = safe_tx_hash[2:]
            self.context.logger.info(f"Hash of the Safe transaction: {safe_tx_hash}")

            payload_string = hash_payload_to_hex(
                safe_tx_hash=safe_tx_hash,
                ether_value=0,
                safe_tx_gas=strategy["safe_tx_gas"]["enter"],
                to_address=self.synchronized_data.multisend_contract_address,
                data=bytes.fromhex(multisend_data),
                operation=SafeOperation.DELEGATE_CALL.value,
            )

            payload = TransactionHashPayload(
                sender=self.context.agent_address, tx_hash=payload_string
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class SwapBackTransactionHashBehaviour(LiquidityRebalancingBaseBehaviour):
    """Prepare the transaction hash for swapping back assets

    The expected transfers derived from this behaviour are
    Safe         ->  A-Base-pool    : A tokens
    A-Base-pool  ->  Safe           : Base tokens
    Safe         ->  B-Base-pool    : B tokens
    B-Base-pool  ->  Safe           : Base tokens
    """

    behaviour_id = "swap_back_tx_hash"
    matching_round = SwapBackTransactionHashRound

    def async_act(self) -> Generator:  # pylint: disable=too-many-statements
        """
        Do the action.

        Steps:
        - Request the transaction hash for the safe transaction. This is the
          hash that needs to be signed by a threshold of agents.
        - Send the transaction hash as a transaction and wait for it to be mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():

            strategy = json.loads(self.synchronized_data.most_voted_strategy)

            transfers = yield from self.get_tx_result()
            transfers = transfers if transfers else []

            strategy["token_a"]["amount_received"] = parse_tx_token_balance(
                transfer_logs=transfers,
                token_address=strategy["token_a"]["address"],
                source_address=strategy["token_LP"]["address"],
                destination_address=self.synchronized_data.safe_contract_address,
            )
            strategy["token_b"]["amount_received"] = parse_tx_token_balance(
                transfer_logs=transfers,
                token_address=strategy["token_b"]["address"],
                source_address=strategy["token_LP"]["address"],
                destination_address=self.synchronized_data.safe_contract_address,
            )

            # Prepare a uniswap tx list. We should check what token balances we have at this point.
            # It is possible that we don't need to swap. For now let's assume we have just USDT
            # and always swap back to it.
            multi_send_txs = []

            # Swap first token back
            swap_tx_data = yield from self.get_swap_data(  # nosec
                strategy=strategy, token="token_a", is_swap_back=True
            )
            if swap_tx_data:
                multi_send_txs.append(swap_tx_data)

            # Swap second token back
            swap_tx_data = yield from self.get_swap_data(  # nosec
                strategy=strategy, token="token_b", is_swap_back=True
            )
            if swap_tx_data:
                multi_send_txs.append(swap_tx_data)

            # Remove allowance for base token
            if (
                not strategy["token_base"]["is_native"]
                and "remove_allowance" in strategy["token_base"]
            ):
                allowance_base_data = yield from self.get_allowance_data(
                    token_address=strategy["token_base"]["address"],
                    value=strategy["token_base"]["remove_allowance"],
                )

                multi_send_txs.append(allowance_base_data)

            # Remove allowance for the first token
            if (
                not strategy["token_a"]["is_native"]
                and "remove_allowance" in strategy["token_a"]
            ):
                allowance_base_data = yield from self.get_allowance_data(
                    token_address=strategy["token_a"]["address"],
                    value=strategy["token_a"]["remove_allowance"],
                )

                multi_send_txs.append(allowance_base_data)

            # Remove allowance for the second token
            if (
                not strategy["token_b"]["is_native"]
                and "remove_allowance" in strategy["token_b"]
            ):
                allowance_b_data = yield from self.get_allowance_data(
                    token_address=strategy["token_b"]["address"],
                    value=strategy["token_b"]["remove_allowance"],
                )

                multi_send_txs.append(allowance_b_data)

            # Get the tx list data from multisend contract
            contract_api_msg = yield from self.get_contract_api_response(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                contract_address=self.synchronized_data.safe_contract_address,
                contract_id=str(MultiSendContract.contract_id),
                contract_callable="get_tx_data",
                multi_send_txs=multi_send_txs,
            )
            multisend_data = cast(str, contract_api_msg.raw_transaction.body["data"])
            multisend_data = multisend_data[2:]
            self.context.logger.info(f"Multisend data: {multisend_data}")
            # Get the tx hash from Gnosis Safe contract
            contract_api_msg = yield from self.get_contract_api_response(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                contract_address=self.synchronized_data.safe_contract_address,
                contract_id=str(GnosisSafeContract.contract_id),
                contract_callable="get_raw_safe_transaction_hash",
                to_address=self.synchronized_data.multisend_contract_address,
                value=0,
                data=bytes.fromhex(multisend_data),
                operation=SafeOperation.DELEGATE_CALL.value,
                safe_tx_gas=strategy["safe_tx_gas"]["swap_back"],
                safe_nonce=strategy["safe_nonce"],
            )
            safe_tx_hash = cast(str, contract_api_msg.raw_transaction.body["tx_hash"])
            safe_tx_hash = safe_tx_hash[2:]
            self.context.logger.info(f"Hash of the Safe transaction: {safe_tx_hash}")

            payload_string = hash_payload_to_hex(
                safe_tx_hash=safe_tx_hash,
                ether_value=0,
                safe_tx_gas=strategy["safe_tx_gas"]["enter"],
                to_address=self.synchronized_data.multisend_contract_address,
                data=bytes.fromhex(multisend_data),
                operation=SafeOperation.DELEGATE_CALL.value,
            )

            payload = TransactionHashPayload(
                sender=self.context.agent_address, tx_hash=payload_string
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class LiquidityRebalancingConsensusBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the rebalancing behaviour."""

    initial_behaviour_cls = StrategyEvaluationBehaviour
    abci_app_cls = LiquidityRebalancingAbciApp  # type: ignore
    behaviours: Set[Type[BaseBehaviour]] = {  # type: ignore
        StrategyEvaluationBehaviour,  # type: ignore
        SleepBehaviour,  # type: ignore
        EnterPoolTransactionHashBehaviour,  # type: ignore
        ExitPoolTransactionHashBehaviour,  # type: ignore
        SwapBackTransactionHashBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the `tools/general.py` module of the skill."""
import json
from typing import Any, Dict, Optional, Union

import pytest
from _pytest.monkeypatch import MonkeyPatch

from packages.valory.skills.apy_estimation_abci.tools.queries import (
    block_from_timestamp_q,
    eth_price_usd_q,
    finalize_q,
    latest_block,
    pairs_q,
    top_n_pairs_q,
)


eth_q_parameterization = pytest.mark.parametrize(
    (
        "bundle_id",
        "block",
        "expected",
    ),
    [
        (
            0,
            0,
            """
            {
                bundles(
                    first: 1,
                    block: {number: 0},
                    where: {
                        id: 0
                    }
                )
                {ethPrice}
            }
            """,
        ),
        (
            0,
            None,
            """
            {
                bundles(
                    first: 1,
                    where: {
                        id: 0
                    }
                )
                {ethPrice}
            }
            """,
        ),
    ],
)


def identity(arg: Any) -> Any:
    """Define an identity function."""
    return arg


class TestQueries:
    """Tests for `Queries`."""

    @staticmethod
    def test_finalize_q() -> None:
        """Test `finalize_q`."""
        # Test result.
        test_string = "test_string: {test: value}"
        actual = finalize_q(test_string)
        expected: Union[Dict[str, str], bytes] = {"query": test_string}
        expected = json.dumps(expected).encode("utf-8")
        assert actual == expected

    @staticmethod
    @eth_q_parameterization
    def test_eth_price_usd_q(
        bundle_id: int, block: Optional[int], expected: str, monkeypatch: MonkeyPatch
    ) -> None:
        """Test `eth_price_usd_q`."""
        monkeypatch.setattr(
            "packages.valory.skills.apy_estimation_abci.tools.queries.finalize_q",
            identity,
        )
        actual = eth_price_usd_q(bundle_id, block)
        assert actual.split() == expected.split()

    @staticmethod
    def test_block_from_timestamp_q(monkeypatch: MonkeyPatch) -> None:
        """Test `block_from_timestamp_q`."""
        monkeypatch.setattr(
            "packages.valory.skills.apy_estimation_abci.tools.queries.finalize_q",
            identity,
        )
        actual = block_from_timestamp_q(100)
        expected = """
                   {
                   blocks(
                   first:
                   1, orderBy: timestamp, orderDirection: asc, where:
                       {
                           timestamp_gte: 100, timestamp_lte: 700
                       }
                       )
                       {
                           timestamp
                           number
                       }
                       }
                   """
        assert actual.split() == expected.split()

    @staticmethod
    def test_latest_block(monkeypatch: MonkeyPatch) -> None:
        """Test `latest_block`."""
        monkeypatch.setattr(
            "packages.valory.skills.apy_estimation_abci.tools.queries.finalize_q",
            identity,
        )
        actual = latest_block()
        expected = """
                   {
                        blocks(
                            first: 1,
                            orderBy: timestamp,
                            orderDirection: desc
                        )
                        {
                            timestamp
                            number
                        }
                    }
                   """
        assert actual.split() == expected.split()

    @staticmethod
    def test_top_n_pairs_q(monkeypatch: MonkeyPatch) -> None:
        """Test `top_n_pairs_q`."""
        monkeypatch.setattr(
            "packages.valory.skills.apy_estimation_abci.tools.queries.finalize_q",
            identity,
        )
        actual = top_n_pairs_q(0)
        expected = """
                   {
                       pairs(
                           first: 0, orderBy: trackedReserveETH,
                       orderDirection: desc
                       )
                       {id}
                   }
                   """
        assert actual.split() == expected.split()

    @staticmethod
    def test_pairs_q(monkeypatch: MonkeyPatch) -> None:
        """Test `pairs_q`."""
        monkeypatch.setattr(
            "packages.valory.skills.apy_estimation_abci.tools.queries.finalize_q",
            identity,
        )
        actual = pairs_q(0, ["x0", "x1", "x2"])
        expected = """
                   {
                       pairs(
                           where: {id_in: ["x0","x1","x2"]},
                           block: {number: 0}
                       ) {
                           id
                           token0 {
                               id
                               symbol
                               name
                           }
                           token1 {
                               id
                               symbol
                               name
                           }
                           reserve0
                           reserve1
                           totalSupply
                           reserveETH
                           reserveUSD
                           trackedReserveETH
                           token0Price
                           token1Price
                           volumeToken0
                           volumeToken1
                           volumeUSD
                           untrackedVolumeUSD
                           txCount
                           createdAtTimestamp
                           createdAtBlockNumber
                           liquidityProviderCount
                       }
                   }
                   """
        assert actual.split() == expected.split()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/counter skill."""
import logging
import struct
from pathlib import Path
from typing import Any, Tuple, cast
from unittest.mock import patch

from aea.test_tools.test_skill import BaseSkillTestCase

from packages.valory.protocols.abci.custom_types import (
    CheckTxType,
    CheckTxTypeEnum,
    Events,
    ProofOps,
    ValidatorUpdates,
)
from packages.valory.protocols.abci.message import AbciMessage
from packages.valory.skills.counter.dialogues import AbciDialogue, AbciDialogues
from packages.valory.skills.counter.handlers import ABCICounterHandler

from tests.conftest import ROOT_DIR


OK_CODE = 0
ERROR_CODE = 1


class TestCounterHandler(BaseSkillTestCase):
    """Test ABCICounterHandler methods."""

    path_to_skill = Path(ROOT_DIR, "packages", "valory", "skills", "counter")
    abci_counter_handler: ABCICounterHandler
    logger: logging.Logger
    abci_dialogues: AbciDialogues

    @classmethod
    def setup(cls, **kwargs: Any) -> None:
        """Setup the test class."""
        super().setup()
        cls.abci_counter_handler = cast(
            ABCICounterHandler, cls._skill.skill_context.handlers.abci
        )
        cls.logger = cls._skill.skill_context.logger

        cls.abci_dialogues = cast(
            AbciDialogues, cls._skill.skill_context.abci_dialogues
        )

    def test_setup(self) -> None:
        """Test the setup method of the echo handler."""
        with patch.object(self.logger, "log") as mock_logger:
            self.abci_counter_handler.setup()
            mock_logger.assert_any_call(
                logging.DEBUG, "ABCI Handler: setup method called."
            )

        # after
        self.assert_quantity_in_outbox(0)

    def test_teardown(self) -> None:
        """Test the teardown method of the echo handler."""
        with patch.object(self.logger, "log") as mock_logger:
            self.abci_counter_handler.teardown()
            mock_logger.assert_any_call(
                logging.DEBUG, "ABCI Handler: teardown method called."
            )

        # after
        self.assert_quantity_in_outbox(0)

    def test_info(
        self,
    ) -> None:
        """Test ABCICounterHandler.info method."""

        abci_message, abci_dialogue = cast(
            Tuple[AbciMessage, AbciDialogue],
            self.abci_dialogues.create(
                performative=AbciMessage.Performative.REQUEST_INFO,
                counterparty="address",
                version="",
                block_version=0,
                p2p_version=0,
            ),
        )

        reply = self.abci_counter_handler.info(abci_message, abci_dialogue)
        has_attr, msg = self.message_has_attributes(
            actual_message=reply,
            message_type=AbciMessage,
            performative=AbciMessage.Performative.RESPONSE_INFO,
            info_data="",
            app_version=0,
            last_block_height=0,
            last_block_app_hash=b"",
        )
        assert has_attr, msg

    def test_init_chain(
        self,
    ) -> None:
        """Test ABCICounterHandler.init_chain method."""

        abci_message, abci_dialogue = cast(
            Tuple[AbciMessage, AbciDialogue],
            self.abci_dialogues.create(
                performative=AbciMessage.Performative.REQUEST_INIT_CHAIN,
                counterparty="address",
                version="",
                block_version=0,
                p2p_version=0,
            ),
        )

        reply = self.abci_counter_handler.init_chain(abci_message, abci_dialogue)
        has_attr, msg = self.message_has_attributes(
            actual_message=reply,
            message_type=AbciMessage,
            performative=AbciMessage.Performative.RESPONSE_INIT_CHAIN,
            validators=ValidatorUpdates([]),
            app_hash=b"",
        )
        assert has_attr, msg

    def test_query(
        self,
    ) -> None:
        """Test ABCICounterHandler.query method."""

        abci_message, abci_dialogue = cast(
            Tuple[AbciMessage, AbciDialogue],
            self.abci_dialogues.create(
                performative=AbciMessage.Performative.REQUEST_QUERY,
                counterparty="address",
                version="",
                query_data=b"",
                path="/",
                height=1,
                prove=True,
            ),
        )

        reply = self.abci_counter_handler.query(abci_message, abci_dialogue)
        has_attr, msg = self.message_has_attributes(
            actual_message=reply,
            performative=AbciMessage.Performative.RESPONSE_QUERY,
            message_type=AbciMessage,
            code=OK_CODE,
            log="value: 0",
            info="",
            index=0,
            key=b"",
            value=b"\x00\x00\x00\x00",
            proof_ops=ProofOps([]),
            height=self.abci_counter_handler.last_block_height,
            codespace="",
        )
        assert has_attr, msg

    def test_check_tx(
        self,
    ) -> None:
        """Test ABCICounterHandler. method."""

        tx = 10

        abci_message, abci_dialogue = cast(
            Tuple[AbciMessage, AbciDialogue],
            self.abci_dialogues.create(
                performative=AbciMessage.Performative.REQUEST_CHECK_TX,
                counterparty="address",
                version="",
                tx=tx.to_bytes(4, "big"),
                type=CheckTxType(CheckTxTypeEnum.NEW),
            ),
        )

        reply = self.abci_counter_handler.check_tx(abci_message, abci_dialogue)
        has_attr, msg = self.message_has_attributes(
            actual_message=reply,
            performative=AbciMessage.Performative.RESPONSE_CHECK_TX,
            message_type=AbciMessage,
            code=ERROR_CODE,
            log="invalid transaction.",
            info="ERROR: the next count must be a unitary increment.",
        )
        assert has_attr, msg

        tx = self.abci_counter_handler.tx_count + 1

        abci_message, abci_dialogue = cast(
            Tuple[AbciMessage, AbciDialogue],
            self.abci_dialogues.create(
                performative=AbciMessage.Performative.REQUEST_CHECK_TX,
                counterparty="address",
                version="",
                tx=tx.to_bytes(4, "big"),
                type=CheckTxType(CheckTxTypeEnum.NEW),
            ),
        )

        reply = self.abci_counter_handler.check_tx(abci_message, abci_dialogue)
        has_attr, msg = self.message_has_attributes(
            actual_message=reply,
            performative=AbciMessage.Performative.RESPONSE_CHECK_TX,
            message_type=AbciMessage,
            code=OK_CODE,
            log="valid transaction.",
            info="OK: the next count is a unitary increment.",
        )
        assert has_attr, msg

    def test_deliver_tx(
        self,
    ) -> None:
        """Test ABCICounterHandler. method."""

        abci_message, abci_dialogue = cast(
            Tuple[AbciMessage, AbciDialogue],
            self.abci_dialogues.create(
                performative=AbciMessage.Performative.REQUEST_DELIVER_TX,
                counterparty="address",
                version="",
                tx=b"",
            ),
        )

        reply = self.abci_counter_handler.deliver_tx(abci_message, abci_dialogue)
        has_attr, msg = self.message_has_attributes(
            actual_message=reply,
            performative=AbciMessage.Performative.RESPONSE_DELIVER_TX,
            message_type=AbciMessage,
            code=OK_CODE,
            data=b"",
            log="",
            info="",
            gas_wanted=0,
            gas_used=0,
            events=Events([]),
            codespace="",
        )
        assert has_attr, msg

    def test_commit(
        self,
    ) -> None:
        """Test ABCICounterHandler. method."""

        abci_message, abci_dialogue = cast(
            Tuple[AbciMessage, AbciDialogue],
            self.abci_dialogues.create(
                performative=AbciMessage.Performative.REQUEST_COMMIT,
                counterparty="address",
                version="",
            ),
        )

        reply = self.abci_counter_handler.commit(abci_message, abci_dialogue)
        has_attr, msg = self.message_has_attributes(
            actual_message=reply,
            performative=AbciMessage.Performative.RESPONSE_COMMIT,
            message_type=AbciMessage,
            data=struct.pack(">Q", self.abci_counter_handler.tx_count),
            retain_height=0,
        )
        assert has_attr, msg



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Serialization module for abci protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import Any, Dict, cast

from aea.mail.base_pb2 import DialogueMessage
from aea.mail.base_pb2 import Message as ProtobufMessage
from aea.protocols.base import Message, Serializer

from packages.valory.protocols.abci import abci_pb2
from packages.valory.protocols.abci.custom_types import (
    CheckTxType,
    ConsensusParams,
    Events,
    Evidences,
    Header,
    LastCommitInfo,
    ProofOps,
    Result,
    SnapShots,
    Snapshot,
    Timestamp,
    ValidatorUpdates,
)
from packages.valory.protocols.abci.message import AbciMessage


class AbciSerializer(Serializer):
    """Serialization for the 'abci' protocol."""

    @staticmethod
    def encode(msg: Message) -> bytes:
        """
        Encode a 'Abci' message into bytes.

        :param msg: the message object.
        :return: the bytes.
        """
        msg = cast(AbciMessage, msg)
        message_pb = ProtobufMessage()
        dialogue_message_pb = DialogueMessage()
        abci_msg = abci_pb2.AbciMessage()

        dialogue_message_pb.message_id = msg.message_id
        dialogue_reference = msg.dialogue_reference
        dialogue_message_pb.dialogue_starter_reference = dialogue_reference[0]
        dialogue_message_pb.dialogue_responder_reference = dialogue_reference[1]
        dialogue_message_pb.target = msg.target

        performative_id = msg.performative
        if performative_id == AbciMessage.Performative.REQUEST_ECHO:
            performative = abci_pb2.AbciMessage.Request_Echo_Performative()  # type: ignore
            message = msg.message
            performative.message = message
            abci_msg.request_echo.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_FLUSH:
            performative = abci_pb2.AbciMessage.Request_Flush_Performative()  # type: ignore
            abci_msg.request_flush.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_INFO:
            performative = abci_pb2.AbciMessage.Request_Info_Performative()  # type: ignore
            version = msg.version
            performative.version = version
            block_version = msg.block_version
            performative.block_version = block_version
            p2p_version = msg.p2p_version
            performative.p2p_version = p2p_version
            abci_msg.request_info.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_SET_OPTION:
            performative = abci_pb2.AbciMessage.Request_Set_Option_Performative()  # type: ignore
            option_key = msg.option_key
            performative.option_key = option_key
            option_value = msg.option_value
            performative.option_value = option_value
            abci_msg.request_set_option.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_INIT_CHAIN:
            performative = abci_pb2.AbciMessage.Request_Init_Chain_Performative()  # type: ignore
            time = msg.time
            Timestamp.encode(performative.time, time)
            chain_id = msg.chain_id
            performative.chain_id = chain_id
            if msg.is_set("consensus_params"):
                performative.consensus_params_is_set = True
                consensus_params = msg.consensus_params
                ConsensusParams.encode(performative.consensus_params, consensus_params)
            validators = msg.validators
            ValidatorUpdates.encode(performative.validators, validators)
            app_state_bytes = msg.app_state_bytes
            performative.app_state_bytes = app_state_bytes
            initial_height = msg.initial_height
            performative.initial_height = initial_height
            abci_msg.request_init_chain.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_QUERY:
            performative = abci_pb2.AbciMessage.Request_Query_Performative()  # type: ignore
            query_data = msg.query_data
            performative.query_data = query_data
            path = msg.path
            performative.path = path
            height = msg.height
            performative.height = height
            prove = msg.prove
            performative.prove = prove
            abci_msg.request_query.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_BEGIN_BLOCK:
            performative = abci_pb2.AbciMessage.Request_Begin_Block_Performative()  # type: ignore
            hash = msg.hash
            performative.hash = hash
            header = msg.header
            Header.encode(performative.header, header)
            last_commit_info = msg.last_commit_info
            LastCommitInfo.encode(performative.last_commit_info, last_commit_info)
            byzantine_validators = msg.byzantine_validators
            Evidences.encode(performative.byzantine_validators, byzantine_validators)
            abci_msg.request_begin_block.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_CHECK_TX:
            performative = abci_pb2.AbciMessage.Request_Check_Tx_Performative()  # type: ignore
            tx = msg.tx
            performative.tx = tx
            type = msg.type
            CheckTxType.encode(performative.type, type)
            abci_msg.request_check_tx.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_DELIVER_TX:
            performative = abci_pb2.AbciMessage.Request_Deliver_Tx_Performative()  # type: ignore
            tx = msg.tx
            performative.tx = tx
            abci_msg.request_deliver_tx.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_END_BLOCK:
            performative = abci_pb2.AbciMessage.Request_End_Block_Performative()  # type: ignore
            height = msg.height
            performative.height = height
            abci_msg.request_end_block.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_COMMIT:
            performative = abci_pb2.AbciMessage.Request_Commit_Performative()  # type: ignore
            abci_msg.request_commit.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_LIST_SNAPSHOTS:
            performative = abci_pb2.AbciMessage.Request_List_Snapshots_Performative()  # type: ignore
            abci_msg.request_list_snapshots.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_OFFER_SNAPSHOT:
            performative = abci_pb2.AbciMessage.Request_Offer_Snapshot_Performative()  # type: ignore
            snapshot = msg.snapshot
            Snapshot.encode(performative.snapshot, snapshot)
            app_hash = msg.app_hash
            performative.app_hash = app_hash
            abci_msg.request_offer_snapshot.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_LOAD_SNAPSHOT_CHUNK:
            performative = abci_pb2.AbciMessage.Request_Load_Snapshot_Chunk_Performative()  # type: ignore
            height = msg.height
            performative.height = height
            format = msg.format
            performative.format = format
            chunk_index = msg.chunk_index
            performative.chunk_index = chunk_index
            abci_msg.request_load_snapshot_chunk.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.REQUEST_APPLY_SNAPSHOT_CHUNK:
            performative = abci_pb2.AbciMessage.Request_Apply_Snapshot_Chunk_Performative()  # type: ignore
            index = msg.index
            performative.index = index
            chunk = msg.chunk
            performative.chunk = chunk
            chunk_sender = msg.chunk_sender
            performative.chunk_sender = chunk_sender
            abci_msg.request_apply_snapshot_chunk.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_EXCEPTION:
            performative = abci_pb2.AbciMessage.Response_Exception_Performative()  # type: ignore
            error = msg.error
            performative.error = error
            abci_msg.response_exception.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_ECHO:
            performative = abci_pb2.AbciMessage.Response_Echo_Performative()  # type: ignore
            message = msg.message
            performative.message = message
            abci_msg.response_echo.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_FLUSH:
            performative = abci_pb2.AbciMessage.Response_Flush_Performative()  # type: ignore
            abci_msg.response_flush.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_INFO:
            performative = abci_pb2.AbciMessage.Response_Info_Performative()  # type: ignore
            info_data = msg.info_data
            performative.info_data = info_data
            version = msg.version
            performative.version = version
            app_version = msg.app_version
            performative.app_version = app_version
            last_block_height = msg.last_block_height
            performative.last_block_height = last_block_height
            last_block_app_hash = msg.last_block_app_hash
            performative.last_block_app_hash = last_block_app_hash
            abci_msg.response_info.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_SET_OPTION:
            performative = abci_pb2.AbciMessage.Response_Set_Option_Performative()  # type: ignore
            code = msg.code
            performative.code = code
            log = msg.log
            performative.log = log
            info = msg.info
            performative.info = info
            abci_msg.response_set_option.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_INIT_CHAIN:
            performative = abci_pb2.AbciMessage.Response_Init_Chain_Performative()  # type: ignore
            if msg.is_set("consensus_params"):
                performative.consensus_params_is_set = True
                consensus_params = msg.consensus_params
                ConsensusParams.encode(performative.consensus_params, consensus_params)
            validators = msg.validators
            ValidatorUpdates.encode(performative.validators, validators)
            app_hash = msg.app_hash
            performative.app_hash = app_hash
            abci_msg.response_init_chain.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_QUERY:
            performative = abci_pb2.AbciMessage.Response_Query_Performative()  # type: ignore
            code = msg.code
            performative.code = code
            log = msg.log
            performative.log = log
            info = msg.info
            performative.info = info
            index = msg.index
            performative.index = index
            key = msg.key
            performative.key = key
            value = msg.value
            performative.value = value
            proof_ops = msg.proof_ops
            ProofOps.encode(performative.proof_ops, proof_ops)
            height = msg.height
            performative.height = height
            codespace = msg.codespace
            performative.codespace = codespace
            abci_msg.response_query.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_BEGIN_BLOCK:
            performative = abci_pb2.AbciMessage.Response_Begin_Block_Performative()  # type: ignore
            events = msg.events
            Events.encode(performative.events, events)
            abci_msg.response_begin_block.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_CHECK_TX:
            performative = abci_pb2.AbciMessage.Response_Check_Tx_Performative()  # type: ignore
            code = msg.code
            performative.code = code
            data = msg.data
            performative.data = data
            log = msg.log
            performative.log = log
            info = msg.info
            performative.info = info
            gas_wanted = msg.gas_wanted
            performative.gas_wanted = gas_wanted
            gas_used = msg.gas_used
            performative.gas_used = gas_used
            events = msg.events
            Events.encode(performative.events, events)
            codespace = msg.codespace
            performative.codespace = codespace
            abci_msg.response_check_tx.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_DELIVER_TX:
            performative = abci_pb2.AbciMessage.Response_Deliver_Tx_Performative()  # type: ignore
            code = msg.code
            performative.code = code
            data = msg.data
            performative.data = data
            log = msg.log
            performative.log = log
            info = msg.info
            performative.info = info
            gas_wanted = msg.gas_wanted
            performative.gas_wanted = gas_wanted
            gas_used = msg.gas_used
            performative.gas_used = gas_used
            events = msg.events
            Events.encode(performative.events, events)
            codespace = msg.codespace
            performative.codespace = codespace
            abci_msg.response_deliver_tx.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_END_BLOCK:
            performative = abci_pb2.AbciMessage.Response_End_Block_Performative()  # type: ignore
            validator_updates = msg.validator_updates
            ValidatorUpdates.encode(performative.validator_updates, validator_updates)
            if msg.is_set("consensus_param_updates"):
                performative.consensus_param_updates_is_set = True
                consensus_param_updates = msg.consensus_param_updates
                ConsensusParams.encode(
                    performative.consensus_param_updates, consensus_param_updates
                )
            events = msg.events
            Events.encode(performative.events, events)
            abci_msg.response_end_block.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_COMMIT:
            performative = abci_pb2.AbciMessage.Response_Commit_Performative()  # type: ignore
            data = msg.data
            performative.data = data
            retain_height = msg.retain_height
            performative.retain_height = retain_height
            abci_msg.response_commit.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_LIST_SNAPSHOTS:
            performative = abci_pb2.AbciMessage.Response_List_Snapshots_Performative()  # type: ignore
            snapshots = msg.snapshots
            SnapShots.encode(performative.snapshots, snapshots)
            abci_msg.response_list_snapshots.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_OFFER_SNAPSHOT:
            performative = abci_pb2.AbciMessage.Response_Offer_Snapshot_Performative()  # type: ignore
            result = msg.result
            Result.encode(performative.result, result)
            abci_msg.response_offer_snapshot.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_LOAD_SNAPSHOT_CHUNK:
            performative = abci_pb2.AbciMessage.Response_Load_Snapshot_Chunk_Performative()  # type: ignore
            chunk = msg.chunk
            performative.chunk = chunk
            abci_msg.response_load_snapshot_chunk.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.RESPONSE_APPLY_SNAPSHOT_CHUNK:
            performative = abci_pb2.AbciMessage.Response_Apply_Snapshot_Chunk_Performative()  # type: ignore
            result = msg.result
            Result.encode(performative.result, result)
            refetch_chunks = msg.refetch_chunks
            performative.refetch_chunks.extend(refetch_chunks)
            reject_senders = msg.reject_senders
            performative.reject_senders.extend(reject_senders)
            abci_msg.response_apply_snapshot_chunk.CopyFrom(performative)
        elif performative_id == AbciMessage.Performative.DUMMY:
            performative = abci_pb2.AbciMessage.Dummy_Performative()  # type: ignore
            dummy_consensus_params = msg.dummy_consensus_params
            ConsensusParams.encode(
                performative.dummy_consensus_params, dummy_consensus_params
            )
            abci_msg.dummy.CopyFrom(performative)
        else:
            raise ValueError("Performative not valid: {}".format(performative_id))

        dialogue_message_pb.content = abci_msg.SerializeToString()

        message_pb.dialogue_message.CopyFrom(dialogue_message_pb)
        message_bytes = message_pb.SerializeToString()
        return message_bytes

    @staticmethod
    def decode(obj: bytes) -> Message:
        """
        Decode bytes into a 'Abci' message.

        :param obj: the bytes object.
        :return: the 'Abci' message.
        """
        message_pb = ProtobufMessage()
        abci_pb = abci_pb2.AbciMessage()
        message_pb.ParseFromString(obj)
        message_id = message_pb.dialogue_message.message_id
        dialogue_reference = (
            message_pb.dialogue_message.dialogue_starter_reference,
            message_pb.dialogue_message.dialogue_responder_reference,
        )
        target = message_pb.dialogue_message.target

        abci_pb.ParseFromString(message_pb.dialogue_message.content)
        performative = abci_pb.WhichOneof("performative")
        performative_id = AbciMessage.Performative(str(performative))
        performative_content = dict()  # type: Dict[str, Any]
        if performative_id == AbciMessage.Performative.REQUEST_ECHO:
            message = abci_pb.request_echo.message
            performative_content["message"] = message
        elif performative_id == AbciMessage.Performative.REQUEST_FLUSH:
            pass
        elif performative_id == AbciMessage.Performative.REQUEST_INFO:
            version = abci_pb.request_info.version
            performative_content["version"] = version
            block_version = abci_pb.request_info.block_version
            performative_content["block_version"] = block_version
            p2p_version = abci_pb.request_info.p2p_version
            performative_content["p2p_version"] = p2p_version
        elif performative_id == AbciMessage.Performative.REQUEST_SET_OPTION:
            option_key = abci_pb.request_set_option.option_key
            performative_content["option_key"] = option_key
            option_value = abci_pb.request_set_option.option_value
            performative_content["option_value"] = option_value
        elif performative_id == AbciMessage.Performative.REQUEST_INIT_CHAIN:
            pb2_time = abci_pb.request_init_chain.time
            time = Timestamp.decode(pb2_time)
            performative_content["time"] = time
            chain_id = abci_pb.request_init_chain.chain_id
            performative_content["chain_id"] = chain_id
            if abci_pb.request_init_chain.consensus_params_is_set:
                pb2_consensus_params = abci_pb.request_init_chain.consensus_params
                consensus_params = ConsensusParams.decode(pb2_consensus_params)
                performative_content["consensus_params"] = consensus_params
            pb2_validators = abci_pb.request_init_chain.validators
            validators = ValidatorUpdates.decode(pb2_validators)
            performative_content["validators"] = validators
            app_state_bytes = abci_pb.request_init_chain.app_state_bytes
            performative_content["app_state_bytes"] = app_state_bytes
            initial_height = abci_pb.request_init_chain.initial_height
            performative_content["initial_height"] = initial_height
        elif performative_id == AbciMessage.Performative.REQUEST_QUERY:
            query_data = abci_pb.request_query.query_data
            performative_content["query_data"] = query_data
            path = abci_pb.request_query.path
            performative_content["path"] = path
            height = abci_pb.request_query.height
            performative_content["height"] = height
            prove = abci_pb.request_query.prove
            performative_content["prove"] = prove
        elif performative_id == AbciMessage.Performative.REQUEST_BEGIN_BLOCK:
            hash = abci_pb.request_begin_block.hash
            performative_content["hash"] = hash
            pb2_header = abci_pb.request_begin_block.header
            header = Header.decode(pb2_header)
            performative_content["header"] = header
            pb2_last_commit_info = abci_pb.request_begin_block.last_commit_info
            last_commit_info = LastCommitInfo.decode(pb2_last_commit_info)
            performative_content["last_commit_info"] = last_commit_info
            pb2_byzantine_validators = abci_pb.request_begin_block.byzantine_validators
            byzantine_validators = Evidences.decode(pb2_byzantine_validators)
            performative_content["byzantine_validators"] = byzantine_validators
        elif performative_id == AbciMessage.Performative.REQUEST_CHECK_TX:
            tx = abci_pb.request_check_tx.tx
            performative_content["tx"] = tx
            pb2_type = abci_pb.request_check_tx.type
            type = CheckTxType.decode(pb2_type)
            performative_content["type"] = type
        elif performative_id == AbciMessage.Performative.REQUEST_DELIVER_TX:
            tx = abci_pb.request_deliver_tx.tx
            performative_content["tx"] = tx
        elif performative_id == AbciMessage.Performative.REQUEST_END_BLOCK:
            height = abci_pb.request_end_block.height
            performative_content["height"] = height
        elif performative_id == AbciMessage.Performative.REQUEST_COMMIT:
            pass
        elif performative_id == AbciMessage.Performative.REQUEST_LIST_SNAPSHOTS:
            pass
        elif performative_id == AbciMessage.Performative.REQUEST_OFFER_SNAPSHOT:
            pb2_snapshot = abci_pb.request_offer_snapshot.snapshot
            snapshot = Snapshot.decode(pb2_snapshot)
            performative_content["snapshot"] = snapshot
            app_hash = abci_pb.request_offer_snapshot.app_hash
            performative_content["app_hash"] = app_hash
        elif performative_id == AbciMessage.Performative.REQUEST_LOAD_SNAPSHOT_CHUNK:
            height = abci_pb.request_load_snapshot_chunk.height
            performative_content["height"] = height
            format = abci_pb.request_load_snapshot_chunk.format
            performative_content["format"] = format
            chunk_index = abci_pb.request_load_snapshot_chunk.chunk_index
            performative_content["chunk_index"] = chunk_index
        elif performative_id == AbciMessage.Performative.REQUEST_APPLY_SNAPSHOT_CHUNK:
            index = abci_pb.request_apply_snapshot_chunk.index
            performative_content["index"] = index
            chunk = abci_pb.request_apply_snapshot_chunk.chunk
            performative_content["chunk"] = chunk
            chunk_sender = abci_pb.request_apply_snapshot_chunk.chunk_sender
            performative_content["chunk_sender"] = chunk_sender
        elif performative_id == AbciMessage.Performative.RESPONSE_EXCEPTION:
            error = abci_pb.response_exception.error
            performative_content["error"] = error
        elif performative_id == AbciMessage.Performative.RESPONSE_ECHO:
            message = abci_pb.response_echo.message
            performative_content["message"] = message
        elif performative_id == AbciMessage.Performative.RESPONSE_FLUSH:
            pass
        elif performative_id == AbciMessage.Performative.RESPONSE_INFO:
            info_data = abci_pb.response_info.info_data
            performative_content["info_data"] = info_data
            version = abci_pb.response_info.version
            performative_content["version"] = version
            app_version = abci_pb.response_info.app_version
            performative_content["app_version"] = app_version
            last_block_height = abci_pb.response_info.last_block_height
            performative_content["last_block_height"] = last_block_height
            last_block_app_hash = abci_pb.response_info.last_block_app_hash
            performative_content["last_block_app_hash"] = last_block_app_hash
        elif performative_id == AbciMessage.Performative.RESPONSE_SET_OPTION:
            code = abci_pb.response_set_option.code
            performative_content["code"] = code
            log = abci_pb.response_set_option.log
            performative_content["log"] = log
            info = abci_pb.response_set_option.info
            performative_content["info"] = info
        elif performative_id == AbciMessage.Performative.RESPONSE_INIT_CHAIN:
            if abci_pb.response_init_chain.consensus_params_is_set:
                pb2_consensus_params = abci_pb.response_init_chain.consensus_params
                consensus_params = ConsensusParams.decode(pb2_consensus_params)
                performative_content["consensus_params"] = consensus_params
            pb2_validators = abci_pb.response_init_chain.validators
            validators = ValidatorUpdates.decode(pb2_validators)
            performative_content["validators"] = validators
            app_hash = abci_pb.response_init_chain.app_hash
            performative_content["app_hash"] = app_hash
        elif performative_id == AbciMessage.Performative.RESPONSE_QUERY:
            code = abci_pb.response_query.code
            performative_content["code"] = code
            log = abci_pb.response_query.log
            performative_content["log"] = log
            info = abci_pb.response_query.info
            performative_content["info"] = info
            index = abci_pb.response_query.index
            performative_content["index"] = index
            key = abci_pb.response_query.key
            performative_content["key"] = key
            value = abci_pb.response_query.value
            performative_content["value"] = value
            pb2_proof_ops = abci_pb.response_query.proof_ops
            proof_ops = ProofOps.decode(pb2_proof_ops)
            performative_content["proof_ops"] = proof_ops
            height = abci_pb.response_query.height
            performative_content["height"] = height
            codespace = abci_pb.response_query.codespace
            performative_content["codespace"] = codespace
        elif performative_id == AbciMessage.Performative.RESPONSE_BEGIN_BLOCK:
            pb2_events = abci_pb.response_begin_block.events
            events = Events.decode(pb2_events)
            performative_content["events"] = events
        elif performative_id == AbciMessage.Performative.RESPONSE_CHECK_TX:
            code = abci_pb.response_check_tx.code
            performative_content["code"] = code
            data = abci_pb.response_check_tx.data
            performative_content["data"] = data
            log = abci_pb.response_check_tx.log
            performative_content["log"] = log
            info = abci_pb.response_check_tx.info
            performative_content["info"] = info
            gas_wanted = abci_pb.response_check_tx.gas_wanted
            performative_content["gas_wanted"] = gas_wanted
            gas_used = abci_pb.response_check_tx.gas_used
            performative_content["gas_used"] = gas_used
            pb2_events = abci_pb.response_check_tx.events
            events = Events.decode(pb2_events)
            performative_content["events"] = events
            codespace = abci_pb.response_check_tx.codespace
            performative_content["codespace"] = codespace
        elif performative_id == AbciMessage.Performative.RESPONSE_DELIVER_TX:
            code = abci_pb.response_deliver_tx.code
            performative_content["code"] = code
            data = abci_pb.response_deliver_tx.data
            performative_content["data"] = data
            log = abci_pb.response_deliver_tx.log
            performative_content["log"] = log
            info = abci_pb.response_deliver_tx.info
            performative_content["info"] = info
            gas_wanted = abci_pb.response_deliver_tx.gas_wanted
            performative_content["gas_wanted"] = gas_wanted
            gas_used = abci_pb.response_deliver_tx.gas_used
            performative_content["gas_used"] = gas_used
            pb2_events = abci_pb.response_deliver_tx.events
            events = Events.decode(pb2_events)
            performative_content["events"] = events
            codespace = abci_pb.response_deliver_tx.codespace
            performative_content["codespace"] = codespace
        elif performative_id == AbciMessage.Performative.RESPONSE_END_BLOCK:
            pb2_validator_updates = abci_pb.response_end_block.validator_updates
            validator_updates = ValidatorUpdates.decode(pb2_validator_updates)
            performative_content["validator_updates"] = validator_updates
            if abci_pb.response_end_block.consensus_param_updates_is_set:
                pb2_consensus_param_updates = (
                    abci_pb.response_end_block.consensus_param_updates
                )
                consensus_param_updates = ConsensusParams.decode(
                    pb2_consensus_param_updates
                )
                performative_content[
                    "consensus_param_updates"
                ] = consensus_param_updates
            pb2_events = abci_pb.response_end_block.events
            events = Events.decode(pb2_events)
            performative_content["events"] = events
        elif performative_id == AbciMessage.Performative.RESPONSE_COMMIT:
            data = abci_pb.response_commit.data
            performative_content["data"] = data
            retain_height = abci_pb.response_commit.retain_height
            performative_content["retain_height"] = retain_height
        elif performative_id == AbciMessage.Performative.RESPONSE_LIST_SNAPSHOTS:
            pb2_snapshots = abci_pb.response_list_snapshots.snapshots
            snapshots = SnapShots.decode(pb2_snapshots)
            performative_content["snapshots"] = snapshots
        elif performative_id == AbciMessage.Performative.RESPONSE_OFFER_SNAPSHOT:
            pb2_result = abci_pb.response_offer_snapshot.result
            result = Result.decode(pb2_result)
            performative_content["result"] = result
        elif performative_id == AbciMessage.Performative.RESPONSE_LOAD_SNAPSHOT_CHUNK:
            chunk = abci_pb.response_load_snapshot_chunk.chunk
            performative_content["chunk"] = chunk
        elif performative_id == AbciMessage.Performative.RESPONSE_APPLY_SNAPSHOT_CHUNK:
            pb2_result = abci_pb.response_apply_snapshot_chunk.result
            result = Result.decode(pb2_result)
            performative_content["result"] = result
            refetch_chunks = abci_pb.response_apply_snapshot_chunk.refetch_chunks
            refetch_chunks_tuple = tuple(refetch_chunks)
            performative_content["refetch_chunks"] = refetch_chunks_tuple
            reject_senders = abci_pb.response_apply_snapshot_chunk.reject_senders
            reject_senders_tuple = tuple(reject_senders)
            performative_content["reject_senders"] = reject_senders_tuple
        elif performative_id == AbciMessage.Performative.DUMMY:
            pb2_dummy_consensus_params = abci_pb.dummy.dummy_consensus_params
            dummy_consensus_params = ConsensusParams.decode(pb2_dummy_consensus_params)
            performative_content["dummy_consensus_params"] = dummy_consensus_params
        else:
            raise ValueError("Performative not valid: {}.".format(performative_id))

        return AbciMessage(
            message_id=message_id,
            dialogue_reference=dialogue_reference,
            target=target,
            performative=performative,
            **performative_content
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains acn's message definition."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,too-many-branches,not-an-iterable,unidiomatic-typecheck,unsubscriptable-object
import logging
from typing import Any, Set, Tuple, cast

from aea.configurations.base import PublicId
from aea.exceptions import AEAEnforceError, enforce
from aea.protocols.base import Message

from packages.valory.protocols.acn.custom_types import AgentRecord as CustomAgentRecord
from packages.valory.protocols.acn.custom_types import StatusBody as CustomStatusBody


_default_logger = logging.getLogger("aea.packages.valory.protocols.acn.message")

DEFAULT_BODY_SIZE = 4


class AcnMessage(Message):
    """The protocol used for envelope delivery on the ACN."""

    protocol_id = PublicId.from_str("valory/acn:1.1.0")
    protocol_specification_id = PublicId.from_str("aea/acn:1.0.0")

    AgentRecord = CustomAgentRecord

    StatusBody = CustomStatusBody

    class Performative(Message.Performative):
        """Performatives for the acn protocol."""

        AEA_ENVELOPE = "aea_envelope"
        LOOKUP_REQUEST = "lookup_request"
        LOOKUP_RESPONSE = "lookup_response"
        REGISTER = "register"
        STATUS = "status"

        def __str__(self) -> str:
            """Get the string representation."""
            return str(self.value)

    _performatives = {
        "aea_envelope",
        "lookup_request",
        "lookup_response",
        "register",
        "status",
    }
    __slots__: Tuple[str, ...] = tuple()

    class _SlotsCls:
        __slots__ = (
            "agent_address",
            "body",
            "dialogue_reference",
            "envelope",
            "message_id",
            "performative",
            "record",
            "target",
        )

    def __init__(
        self,
        performative: Performative,
        dialogue_reference: Tuple[str, str] = ("", ""),
        message_id: int = 1,
        target: int = 0,
        **kwargs: Any,
    ):
        """
        Initialise an instance of AcnMessage.

        :param message_id: the message id.
        :param dialogue_reference: the dialogue reference.
        :param target: the message target.
        :param performative: the message performative.
        :param **kwargs: extra options.
        """
        super().__init__(
            dialogue_reference=dialogue_reference,
            message_id=message_id,
            target=target,
            performative=AcnMessage.Performative(performative),
            **kwargs,
        )

    @property
    def valid_performatives(self) -> Set[str]:
        """Get valid performatives."""
        return self._performatives

    @property
    def dialogue_reference(self) -> Tuple[str, str]:
        """Get the dialogue_reference of the message."""
        enforce(self.is_set("dialogue_reference"), "dialogue_reference is not set.")
        return cast(Tuple[str, str], self.get("dialogue_reference"))

    @property
    def message_id(self) -> int:
        """Get the message_id of the message."""
        enforce(self.is_set("message_id"), "message_id is not set.")
        return cast(int, self.get("message_id"))

    @property
    def performative(self) -> Performative:  # type: ignore # noqa: F821
        """Get the performative of the message."""
        enforce(self.is_set("performative"), "performative is not set.")
        return cast(AcnMessage.Performative, self.get("performative"))

    @property
    def target(self) -> int:
        """Get the target of the message."""
        enforce(self.is_set("target"), "target is not set.")
        return cast(int, self.get("target"))

    @property
    def agent_address(self) -> str:
        """Get the 'agent_address' content from the message."""
        enforce(self.is_set("agent_address"), "'agent_address' content is not set.")
        return cast(str, self.get("agent_address"))

    @property
    def body(self) -> CustomStatusBody:
        """Get the 'body' content from the message."""
        enforce(self.is_set("body"), "'body' content is not set.")
        return cast(CustomStatusBody, self.get("body"))

    @property
    def envelope(self) -> bytes:
        """Get the 'envelope' content from the message."""
        enforce(self.is_set("envelope"), "'envelope' content is not set.")
        return cast(bytes, self.get("envelope"))

    @property
    def record(self) -> CustomAgentRecord:
        """Get the 'record' content from the message."""
        enforce(self.is_set("record"), "'record' content is not set.")
        return cast(CustomAgentRecord, self.get("record"))

    def _is_consistent(self) -> bool:
        """Check that the message follows the acn protocol."""
        try:
            enforce(
                isinstance(self.dialogue_reference, tuple),
                "Invalid type for 'dialogue_reference'. Expected 'tuple'. Found '{}'.".format(
                    type(self.dialogue_reference)
                ),
            )
            enforce(
                isinstance(self.dialogue_reference[0], str),
                "Invalid type for 'dialogue_reference[0]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[0])
                ),
            )
            enforce(
                isinstance(self.dialogue_reference[1], str),
                "Invalid type for 'dialogue_reference[1]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[1])
                ),
            )
            enforce(
                type(self.message_id) is int,
                "Invalid type for 'message_id'. Expected 'int'. Found '{}'.".format(
                    type(self.message_id)
                ),
            )
            enforce(
                type(self.target) is int,
                "Invalid type for 'target'. Expected 'int'. Found '{}'.".format(
                    type(self.target)
                ),
            )

            # Light Protocol Rule 2
            # Check correct performative
            enforce(
                isinstance(self.performative, AcnMessage.Performative),
                "Invalid 'performative'. Expected either of '{}'. Found '{}'.".format(
                    self.valid_performatives, self.performative
                ),
            )

            # Check correct contents
            actual_nb_of_contents = len(self._body) - DEFAULT_BODY_SIZE
            expected_nb_of_contents = 0
            if self.performative == AcnMessage.Performative.REGISTER:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.record, CustomAgentRecord),
                    "Invalid type for content 'record'. Expected 'AgentRecord'. Found '{}'.".format(
                        type(self.record)
                    ),
                )
            elif self.performative == AcnMessage.Performative.LOOKUP_REQUEST:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.agent_address, str),
                    "Invalid type for content 'agent_address'. Expected 'str'. Found '{}'.".format(
                        type(self.agent_address)
                    ),
                )
            elif self.performative == AcnMessage.Performative.LOOKUP_RESPONSE:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.record, CustomAgentRecord),
                    "Invalid type for content 'record'. Expected 'AgentRecord'. Found '{}'.".format(
                        type(self.record)
                    ),
                )
            elif self.performative == AcnMessage.Performative.AEA_ENVELOPE:
                expected_nb_of_contents = 2
                enforce(
                    isinstance(self.envelope, bytes),
                    "Invalid type for content 'envelope'. Expected 'bytes'. Found '{}'.".format(
                        type(self.envelope)
                    ),
                )
                enforce(
                    isinstance(self.record, CustomAgentRecord),
                    "Invalid type for content 'record'. Expected 'AgentRecord'. Found '{}'.".format(
                        type(self.record)
                    ),
                )
            elif self.performative == AcnMessage.Performative.STATUS:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.body, CustomStatusBody),
                    "Invalid type for content 'body'. Expected 'StatusBody'. Found '{}'.".format(
                        type(self.body)
                    ),
                )

            # Check correct content count
            enforce(
                expected_nb_of_contents == actual_nb_of_contents,
                "Incorrect number of contents. Expected {}. Found {}".format(
                    expected_nb_of_contents, actual_nb_of_contents
                ),
            )

            # Light Protocol Rule 3
            if self.message_id == 1:
                enforce(
                    self.target == 0,
                    "Invalid 'target'. Expected 0 (because 'message_id' is 1). Found {}.".format(
                        self.target
                    ),
                )
        except (AEAEnforceError, ValueError, KeyError) as e:
            _default_logger.error(str(e))
            return False

        return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains all the interaction operations of the behaviours with IPFS."""


import os
from shutil import rmtree
from typing import Any, Optional

from aea_cli_ipfs.ipfs_utils import DownloadError, IPFSTool, NodeError
from ipfshttpclient.exceptions import ErrorResponse

from packages.valory.skills.abstract_round_abci.io.load import (
    CustomLoaderType,
    Loader,
    SupportedFiletype,
    SupportedObjectType,
)
from packages.valory.skills.abstract_round_abci.io.store import CustomStorerType, Storer


class IPFSInteractionError(Exception):
    """A custom exception for IPFS interaction errors."""


class IPFSInteract:
    """Class for interacting with IPFS."""

    def __init__(self, domain: str):
        """Initialize an `IPFSInteract`.

        :param domain: the IPFS domain name.
        """
        try:
            # Create an IPFS tool.
            self.__ipfs_tool = IPFSTool(domain)
            # Check IPFS node.
            self.__ipfs_tool.check_ipfs_node_running()
        except (NodeError, Exception) as e:  # pragma: no cover
            raise IPFSInteractionError(str(e)) from e

    @staticmethod
    def __remove_filepath(filepath: str) -> None:
        """Remove a file or a folder. If filepath is not a file or a folder, an `IPFSInteractionError` is raised."""
        if os.path.isfile(filepath):
            os.remove(filepath)
        elif os.path.isdir(filepath):
            rmtree(filepath)
        else:  # pragma: no cover
            raise IPFSInteractionError(f"`{filepath}` is not an existing filepath!")

    def _send(self, filepath: str) -> str:
        """Send a file to the IPFS node.

        :param filepath: the filepath of the file or folder to send
        :return: the file's hash
        """
        try:
            _, hash_, _ = self.__ipfs_tool.add(filepath)
        except ValueError as e:  # pragma: no cover
            raise IPFSInteractionError(str(e)) from e
        finally:
            self.__remove_filepath(filepath)

        return hash_

    def _download(
        self,
        hash_: str,
        target_dir: str,
        multiple: bool = False,
        filename: Optional[str] = None,
    ) -> str:
        """Download a file from the IPFS node.

        :param hash_: hash of file to download
        :param target_dir: directory to place downloaded file
        :param multiple: whether multiple files are expected to be downloaded. The hash should be a folder's hash.
        :param filename: the original name of the file to download
        :return: the filepath of the downloaded file
        """
        if multiple:
            # IPFS tool creates a folder named as the basename of `target_dir` inside `target_dir`.
            filepath = os.path.join(target_dir, os.path.basename(target_dir))
        elif filename is not None:
            filepath = os.path.join(target_dir, filename)
        else:  # pragma: no cover
            raise IPFSInteractionError(
                "Filename cannot be `None` when uploading a single file!"
            )

        if os.path.exists(target_dir):  # pragma: no cover
            # TODO investigate why sometimes the path exists. It shouldn't, because `_send` removes it.
            self.__remove_filepath(target_dir)

        try:
            self.__ipfs_tool.download(hash_, target_dir)
        except (DownloadError, PermissionError, ErrorResponse) as e:  # pragma: no cover
            raise IPFSInteractionError(str(e)) from e

        return filepath

    def store_and_send(
        self,
        filepath: str,
        obj: SupportedObjectType,
        multiple: bool,
        filetype: Optional[SupportedFiletype] = None,
        custom_storer: Optional[CustomStorerType] = None,
        **kwargs: Any,
    ) -> str:
        """Temporarily store a file locally, in order to send it to IPFS and retrieve a hash, and then delete it."""
        filepath = os.path.normpath(filepath)
        if multiple:
            # Add trailing slash in order to treat path as a folder.
            filepath = os.path.join(filepath, "")
        storer = Storer(filetype, custom_storer, filepath)

        try:
            storer.store(obj, multiple, **kwargs)
        except IOError as e:  # pragma: no cover
            raise IPFSInteractionError(str(e)) from e

        return self._send(filepath)

    def get_and_read(  # pylint: disable=too-many-arguments
        self,
        hash_: str,
        target_dir: str,
        multiple: bool = False,
        filename: Optional[str] = None,
        filetype: Optional[SupportedFiletype] = None,
        custom_loader: CustomLoaderType = None,
    ) -> SupportedObjectType:
        """Get, store and read a file from IPFS."""
        target_dir = os.path.normpath(target_dir)
        filepath = self._download(hash_, target_dir, multiple, filename)
        loader = Loader(filetype, custom_loader)

        try:
            return loader.load(filepath, multiple)
        except IOError as e:  # pragma: no cover
            raise IPFSInteractionError(str(e)) from e



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Base deployments module."""
import abc
import json
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

import yaml
from aea.cli.utils.package_utils import try_get_item_source_path
from aea.configurations.base import (
    ConnectionConfig,
    ContractConfig,
    ProtocolConfig,
    SkillConfig,
)
from aea.configurations.constants import AGENTS, DEFAULT_AEA_CONFIG_FILE

from autonomy.configurations.base import Service
from autonomy.configurations.loader import load_service_config
from autonomy.constants import TENDERMINT_IMAGE_VERSION
from autonomy.deploy.constants import (
    DEFAULT_ENCODING,
    KEY_SCHEMA_ADDRESS,
    KEY_SCHEMA_ENCRYPTED_KEY,
    KEY_SCHEMA_UNENCRYPTED_KEY,
    NETWORKS,
)


ABCI_HOST = "abci{}"
TENDERMINT_NODE = "http://node{}:26657"
TENDERMINT_COM = "http://node{}:8080"
COMPONENT_CONFIGS: Dict = {
    component.package_type.value: component  # type: ignore
    for component in [
        ContractConfig,
        SkillConfig,
        ProtocolConfig,
        ConnectionConfig,
    ]
}


class ServiceSpecification:
    """Class to assist with generating deployments."""

    service: Service
    overrides: List
    packages_dir: Path

    def __init__(
        self,
        service_path: Path,
        keys: Path,
        packages_dir: Path,
        number_of_agents: Optional[int] = None,
        private_keys_password: Optional[str] = None,
    ) -> None:
        """Initialize the Base Deployment."""
        self.private_keys: List = []
        self.private_keys_password = private_keys_password
        self.packages_dir = packages_dir
        self.service = load_service_config(service_path)
        if number_of_agents is not None:
            self.service.number_of_agents = number_of_agents

        self.agent_spec = self.load_agent()
        self.read_keys(keys)

        if self.service.number_of_agents > len(self.private_keys):
            raise ValueError("Number of agents cannot be greater than available keys.")

    def read_keys(self, file_path: Path) -> None:
        """Read in keys from a file on disk."""

        keys = json.loads(file_path.read_text(encoding=DEFAULT_ENCODING))

        key_schema = (
            KEY_SCHEMA_UNENCRYPTED_KEY
            if self.private_keys_password is None
            else KEY_SCHEMA_ENCRYPTED_KEY
        )
        for key in keys:
            for required_key in [KEY_SCHEMA_ADDRESS, key_schema]:
                if required_key not in key.keys():
                    raise ValueError("Key file incorrectly formatted.")
            self.private_keys.append(key[key_schema])

    def process_model_args_overrides(self, agent_n: int) -> Dict:
        """Generates env vars based on model overrides."""
        final_overrides = {}
        for component_configuration_json in self.service.overrides:
            _, overrides = self.service.process_component_section(
                agent_n, component_configuration_json
            )
            final_overrides.update(overrides)
        return final_overrides

    def generate_agents(self) -> List:
        """Generate multiple agent."""
        return [self.generate_agent(i) for i in range(self.service.number_of_agents)]

    def generate_common_vars(self, agent_n: int) -> Dict:
        """Retrieve vars common for valory apps."""
        agent_vars = {
            "ID": agent_n,
            "VALORY_APPLICATION": self.service.agent,
            "ABCI_HOST": ABCI_HOST.format(agent_n),
            "MAX_PARTICIPANTS": self.service.number_of_agents,
            "TENDERMINT_URL": TENDERMINT_NODE.format(agent_n),
            "TENDERMINT_COM_URL": TENDERMINT_COM.format(agent_n),
        }

        if self.private_keys_password is not None:
            agent_vars["AEA_PASSWORD"] = self.private_keys_password
        return agent_vars

    def generate_agent(self, agent_n: int) -> Dict[Any, Any]:
        """Generate next agent."""
        agent_vars = self.generate_common_vars(agent_n)
        if len(self.service.overrides) == 0:
            return agent_vars
        overrides = self.process_model_args_overrides(agent_n)
        agent_vars.update(overrides)
        for var_name, value in agent_vars.items():
            if any([isinstance(value, list), isinstance(value, dict)]):
                agent_vars[var_name] = json.dumps(value)
        return agent_vars

    def load_agent(self) -> List[Dict[str, str]]:
        """Using specified valory app, try to load the aea."""
        aea_project_path = self.locate_agent_from_packages_directory()
        with open(aea_project_path, "r", encoding="utf8") as file:
            aea_json = list(yaml.safe_load_all(file))
        return aea_json

    def locate_agent_from_packages_directory(self, local_registry: bool = True) -> str:
        """Using the deployment id, locate the registry and retrieve the path."""
        if local_registry is False:
            raise ValueError("Remote registry not yet supported, use local!")

        source_path = try_get_item_source_path(
            str(self.packages_dir),
            self.service.agent.author,
            AGENTS,
            self.service.agent.name,
        )
        return str(Path(source_path) / DEFAULT_AEA_CONFIG_FILE)


class BaseDeploymentGenerator:
    """Base Deployment Class."""

    service_spec: ServiceSpecification
    output_name: str
    deployment_type: str
    build_dir: Path
    output: str
    tendermint_job_config: Optional[str]

    def __init__(self, service_spec: ServiceSpecification, build_dir: Path):
        """Initialise with only kwargs."""
        self.network_config = NETWORKS[self.deployment_type][
            cast(str, service_spec.service.network)
        ]
        self.service_spec = service_spec
        self.build_dir = Path(build_dir)
        self.tendermint_job_config: Optional[str] = None

    @abc.abstractmethod
    def generate(
        self,
        image_versions: Dict[str, str],
        dev_mode: bool = False,
    ) -> "BaseDeploymentGenerator":
        """Generate the deployment configuration."""

    @abc.abstractmethod
    def generate_config_tendermint(
        self, image_version: str = TENDERMINT_IMAGE_VERSION
    ) -> "BaseDeploymentGenerator":
        """Generate the deployment configuration."""

    @abc.abstractmethod
    def populate_private_keys(
        self,
    ) -> "BaseDeploymentGenerator":
        """Populate the private keys to the deployment."""

    def get_deployment_network_configuration(
        self, agent_vars: List[Dict[str, Any]]
    ) -> List:
        """Retrieve the appropriate network configuration based on deployment & network."""
        for agent in agent_vars:
            agent.update(self.network_config)
        return agent_vars

    def write_config(self) -> "BaseDeploymentGenerator":
        """Write output to build dir"""

        with open(self.build_dir / self.output_name, "w", encoding="utf8") as f:
            f.write(self.output)
        return self



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the classes required for dialogue management."""

from typing import Any

from aea.protocols.base import Address, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.skills.base import Model

from packages.valory.protocols.abci.dialogues import AbciDialogue as BaseAbciDialogue
from packages.valory.protocols.abci.dialogues import AbciDialogues as BaseAbciDialogues


AbciDialogue = BaseAbciDialogue


class AbciDialogues(Model, BaseAbciDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize dialogues.

        :param kwargs: keyword arguments
        """
        Model.__init__(self, **kwargs)

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return AbciDialogue.Role.CLIENT

        BaseAbciDialogues.__init__(
            self,
            self_address=str(self.skill_id),
            role_from_first_message=role_from_first_message,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Configurations for APY skill's tools tests."""

from typing import List, Union

import pandas as pd
import pytest

from packages.valory.skills.apy_estimation_abci.tools.io import TRANSFORMED_HIST_DTYPES


@pytest.fixture
def transformed_history() -> pd.DataFrame:
    """Generate a test transformed history.

    :return: a test transformed history.
    """
    hist_li: List[Union[int, float, str]] = [10] * 21
    hist_li.extend(["x"] * 4)
    hist_li.insert(23, "X")
    hist_li.append("X")
    hist_li.append("x - x")
    hist_li.append(10.0)
    hist_li.append(100.0)
    hist_li.extend([0.0] * 2)

    transformed = pd.DataFrame([hist_li], [1], TRANSFORMED_HIST_DTYPES.keys())

    return transformed



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the common.py module of the skill."""

import binascii
import json
import re
import time
from pathlib import Path
from typing import Any, Set, Type, cast
from unittest import mock

import pytest
from aea.exceptions import AEAActException
from aea.skills.base import SkillContext

from packages.valory.protocols.contract_api.custom_types import State
from packages.valory.protocols.ledger_api.message import LedgerApiMessage
from packages.valory.skills.abstract_round_abci.base import (
    AbciAppDB,
    BaseSynchronizedData,
)
from packages.valory.skills.abstract_round_abci.behaviour_utils import BaseBehaviour
from packages.valory.skills.abstract_round_abci.common import random_selection
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    VerificationStatus,
)

from tests.conftest import ROOT_DIR
from tests.test_skills.base import FSMBehaviourBaseCase


DRAND_VALUE = {
    "round": 1416669,
    "randomness": "f6be4bf1fa229f22340c1a5b258f809ac4af558200775a67dacb05f0cb258a11",
    "signature": (
        "b44d00516f46da3a503f9559a634869b6dc2e5d839e46ec61a090e3032172954929a5"
        "d9bd7197d7739fe55db770543c71182562bd0ad20922eb4fe6b8a1062ed21df3b68de"
        "44694eb4f20b35262fa9d63aa80ad3f6172dd4d33a663f21179604"
    ),
    "previous_signature": (
        "903c60a4b937a804001032499a855025573040cb86017c38e2b1c3725286756ce8f33"
        "61188789c17336beaf3f9dbf84b0ad3c86add187987a9a0685bc5a303e37b008fba8c"
        "44f02a416480dd117a3ff8b8075b1b7362c58af195573623187463"
    ),
}


class CommonBaseCase(FSMBehaviourBaseCase):
    """Base case for testing PriceEstimation FSMBehaviour."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "valory", "skills", "price_estimation_abci"
    )


class BaseRandomnessBehaviourTest(CommonBaseCase):
    """Test RandomnessBehaviour."""

    randomness_behaviour_class: Type[BaseBehaviour]
    next_behaviour_class: Type[BaseBehaviour]
    done_event: Any

    def test_randomness_behaviour(
        self,
    ) -> None:
        """Test RandomnessBehaviour."""

        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            BaseSynchronizedData(AbciAppDB(initial_data={})),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        self.behaviour.act_wrapper()
        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                headers="",
                version="",
                body=b"",
                url="https://drand.cloudflare.com/public/latest",
            ),
            response_kwargs=dict(
                version="",
                status_code=200,
                status_text="",
                headers="",
                body=json.dumps(DRAND_VALUE).encode("utf-8"),
            ),
        )

        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(self.done_event)

        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id

    def test_invalid_drand_value(
        self,
    ) -> None:
        """Test invalid drand values."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            BaseSynchronizedData(AbciAppDB(initial_data={})),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        self.behaviour.act_wrapper()

        drand_value = DRAND_VALUE.copy()
        drand_value["randomness"] = binascii.hexlify(b"randomness_hex").decode()
        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                headers="",
                version="",
                body=b"",
                url="https://drand.cloudflare.com/public/latest",
            ),
            response_kwargs=dict(
                version="",
                status_code=200,
                status_text="",
                headers="",
                body=json.dumps(drand_value).encode(),
            ),
        )

    def test_invalid_response(
        self,
    ) -> None:
        """Test invalid json response."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            BaseSynchronizedData(AbciAppDB(initial_data={})),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        self.behaviour.act_wrapper()

        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                headers="",
                version="",
                body=b"",
                url="https://drand.cloudflare.com/public/latest",
            ),
            response_kwargs=dict(
                version="", status_code=200, status_text="", headers="", body=b""
            ),
        )
        self.behaviour.act_wrapper()
        time.sleep(1)
        self.behaviour.act_wrapper()

    def test_max_retries_reached_fallback(
        self,
    ) -> None:
        """Test with max retries reached."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            BaseSynchronizedData(AbciAppDB(initial_data={})),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        with mock.patch.object(
            self.behaviour.context.randomness_api,
            "is_retries_exceeded",
            return_value=True,
        ):
            self.behaviour.act_wrapper()
            self.mock_ledger_api_request(
                request_kwargs=dict(
                    performative=LedgerApiMessage.Performative.GET_STATE
                ),
                response_kwargs=dict(
                    performative=LedgerApiMessage.Performative.STATE,
                    state=State(ledger_id="ethereum", body={"hash": "0xa"}),
                ),
            )

            self.behaviour.act_wrapper()
            self.mock_a2a_transaction()
            self._test_done_flag_set()
            self.end_round(self.done_event)

            behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
            assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id

    def test_max_retries_reached_fallback_fail(
        self,
    ) -> None:
        """Test with max retries reached."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            BaseSynchronizedData(AbciAppDB(initial_data={})),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        with mock.patch.object(
            self.behaviour.context.randomness_api,
            "is_retries_exceeded",
            return_value=True,
        ):
            self.behaviour.act_wrapper()
            self.mock_ledger_api_request(
                request_kwargs=dict(
                    performative=LedgerApiMessage.Performative.GET_STATE
                ),
                response_kwargs=dict(
                    performative=LedgerApiMessage.Performative.ERROR,
                    state=State(ledger_id="ethereum", body={}),
                ),
            )

            self.behaviour.act_wrapper()

    def test_max_retries_reached_fallback_fail_case_2(
        self,
    ) -> None:
        """Test with max retries reached."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            BaseSynchronizedData(AbciAppDB(initial_data={})),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        with mock.patch.object(
            self.behaviour.context.randomness_api,
            "is_retries_exceeded",
            return_value=True,
        ):
            self.behaviour.act_wrapper()
            self.mock_ledger_api_request(
                request_kwargs=dict(
                    performative=LedgerApiMessage.Performative.GET_STATE
                ),
                response_kwargs=dict(
                    performative=LedgerApiMessage.Performative.STATE,
                    state=State(ledger_id="ethereum", body={}),
                ),
            )

            self.behaviour.act_wrapper()

    def test_clean_up(
        self,
    ) -> None:
        """Test when `observed` value is none."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.randomness_behaviour_class.behaviour_id,
            BaseSynchronizedData(AbciAppDB(initial_data={})),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.randomness_behaviour_class.behaviour_id
        )
        self.behaviour.context.randomness_api._retries_attempted = 1
        assert self.behaviour.current_behaviour is not None
        self.behaviour.current_behaviour.clean_up()
        assert self.behaviour.context.randomness_api._retries_attempted == 0


class BaseSelectKeeperBehaviourTest(CommonBaseCase):
    """Test SelectKeeperBehaviour."""

    select_keeper_behaviour_class: Type[BaseBehaviour]
    next_behaviour_class: Type[BaseBehaviour]
    done_event: Any
    _synchronized_data: Type[BaseSynchronizedData] = BaseSynchronizedData

    @mock.patch.object(SkillContext, "agent_address", new_callable=mock.PropertyMock)
    @pytest.mark.parametrize(
        "blacklisted_keepers",
        (
            set(),
            {"a_1"},
            {"test_agent_address" + "t" * 24},
            {"a_1" + "t" * 39, "a_2" + "t" * 39, "test_agent_address" + "t" * 24},
        ),
    )
    def test_select_keeper(
        self, agent_address_mock: mock.Mock, blacklisted_keepers: Set[str]
    ) -> None:
        """Test select keeper agent."""
        agent_address_mock.return_value = "test_agent_address" + "t" * 24
        participants = frozenset(
            {self.skill.skill_context.agent_address, "a_1" + "t" * 39, "a_2" + "t" * 39}
        )
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.select_keeper_behaviour_class.behaviour_id,
            synchronized_data=self._synchronized_data(
                AbciAppDB(
                    initial_data=AbciAppDB.data_to_lists(
                        dict(
                            participants=participants,
                            most_voted_randomness="56cbde9e9bbcbdcaf92f183c678eaa5288581f06b1c9c7f884ce911776727688",
                            final_verification_status=VerificationStatus.PENDING,
                            blacklisted_keepers="".join(blacklisted_keepers),
                        )
                    ),
                )
            ),
        )
        assert self.behaviour.current_behaviour is not None
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == self.select_keeper_behaviour_class.behaviour_id
        )

        if (
            self.behaviour.current_behaviour.synchronized_data.participants
            - self.behaviour.current_behaviour.synchronized_data.blacklisted_keepers
        ):
            self.behaviour.act_wrapper()
            self.mock_a2a_transaction()
            self._test_done_flag_set()
            self.end_round(self.done_event)
            assert (
                self.behaviour.current_behaviour.behaviour_id
                == self.next_behaviour_class.behaviour_id
            )
        else:
            with pytest.raises(
                AEAActException,
                match="Cannot continue if all the keepers have been blacklisted!",
            ):
                self.behaviour.act_wrapper()

    def test_select_keeper_preexisting_keeper(
        self,
    ) -> None:
        """Test select keeper agent."""
        participants = frozenset({self.skill.skill_context.agent_address, "a_1", "a_2"})
        preexisting_keeper = next(iter(participants))
        self.fast_forward_to_behaviour(
            behaviour=self.behaviour,
            behaviour_id=self.select_keeper_behaviour_class.behaviour_id,
            synchronized_data=self._synchronized_data(
                AbciAppDB(
                    initial_data=dict(
                        participants=[participants],
                        most_voted_randomness=[
                            "56cbde9e9bbcbdcaf92f183c678eaa5288581f06b1c9c7f884ce911776727688"
                        ],
                        most_voted_keeper_address=[preexisting_keeper],
                    ),
                )
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.select_keeper_behaviour_class.behaviour_id
        )
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(self.done_event)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


def test_random_selection() -> None:
    """Test 'random_selection'"""
    assert random_selection(elements=[0, 1, 2], randomness=0.25) == 0
    assert random_selection(elements=[0, 1, 2], randomness=0.5) == 1
    assert random_selection(elements=[0, 1, 2], randomness=0.75) == 2

    with pytest.raises(
        ValueError, match=re.escape("Randomness should lie in the [0,1) interval")
    ):
        random_selection(elements=[0, 1], randomness=-1)

    with pytest.raises(
        ValueError, match=re.escape("Randomness should lie in the [0,1) interval")
    ):
        random_selection(elements=[0, 1], randomness=1)

    with pytest.raises(
        ValueError, match=re.escape("Randomness should lie in the [0,1) interval")
    ):
        random_selection(elements=[0, 1], randomness=2)

    with pytest.raises(ValueError, match="No elements to randomly select among"):
        random_selection(elements=[], randomness=0.5)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Forecasting operations"""
import warnings
from typing import Any, Dict, Optional, Union, cast

import numpy as np
import pandas as pd
from pmdarima import ARIMA
from pmdarima.metrics import smape
from pmdarima.pipeline import Pipeline
from pmdarima.preprocessing import FourierFeaturizer
from sklearn.metrics import (
    explained_variance_score,
    max_error,
    mean_pinball_loss,
    mean_squared_error,
)


MetricsType = Dict[str, float]
TestReportType = Dict[str, str]
PoolIdToTestReportType = Dict[str, TestReportType]
PoolIdToTrainDataType = Dict[str, np.ndarray]
PoolIdToForecasterType = Dict[str, Pipeline]
HyperParamsType = Dict[str, Any]
PoolToHyperParamsType = Dict[str, HyperParamsType]


def init_forecaster(  # pylint: disable=too-many-arguments
    p: int,
    q: int,
    d: int,
    m: int,
    k: Optional[int] = None,
    maxiter: int = 150,
    suppress_warnings: bool = True,
) -> Pipeline:
    """Initialize a forecasting model.

    :param m: the seasonal periodicity of the endogenous vector, y.
    :param k: the number of sine and cosine terms (each) to include.
        I.e., if k is 2, 4 new features will be generated. k must not exceed m/2,
        which is the default value if not set. The value of k can be selected by minimizing the AIC.
    :param p: the order (number of time lags) of the autoregressive model (AR).
    :param q: the order of the moving-average model (MA).
    :param d: the degree of differencing (the number of times the data have had past values subtracted) (I).
    :param maxiter: the maximum number of function evaluations. Default is 150.
    :param suppress_warnings: many warnings might be thrown inside of `statsmodels` - which is used by `pmdarima` - .
        If suppress_warnings is True, all of these warnings will be squelched. Default is True.
    :return: a `pmdarima` pipeline, consisting of a fourier featurizer and an ARIMA model.
    """
    order = (p, q, d)

    # The Pipeline is deterministic.
    forecaster = Pipeline(
        [
            ("fourier", FourierFeaturizer(m, k)),
            (
                "arima",
                ARIMA(order, maxiter=maxiter, suppress_warnings=suppress_warnings),
            ),
        ]
    )

    return forecaster


def train_forecaster_per_pool(
    y_train: PoolIdToTrainDataType, best_params: PoolToHyperParamsType
) -> PoolIdToForecasterType:
    """Train a forecasting model.

    :param y_train: the training timeseries.
    :param best_params: the best parameters for the forecasters.
    :return: a trained `pmdarima` pipeline per pool, consisting of a fourier featurizer and an ARIMA model.
    """
    forecasters = {}
    for id_, y in y_train.items():
        forecasters[id_.replace(".csv", ".joblib")] = train_forecaster(
            y, **best_params[id_]
        )
    return forecasters


def train_forecaster(y_train: np.ndarray, **kwargs: Any) -> Pipeline:
    """Train a forecasting model.

    :param y_train: the training timeseries.
    :param kwargs: the keyword arguments for the forecaster's training.
    :return: a trained `pmdarima` pipeline, consisting of a fourier featurizer and an ARIMA model.
    """
    forecaster = init_forecaster(**kwargs)
    forecaster.fit(y_train)

    return forecaster


def baseline(t0: float, y_test: np.ndarray) -> np.ndarray:
    """Create the baseline model's 'predictions'.

    Given a timeseries, the baseline model will be "predicting" at each step $t_n$ the value of $t_{n-1}$.

    :param t0: the current timestep, i.e., the last value of the training set.
    :param y_test: the test values.
    :return: the "predictions" of the baseline model.
    """
    # TODO consider walk_forward arg as well and pass a t0 of equal len.
    y_test = np.insert(y_test, 0, t0)

    return y_test[:-1]


def calc_metrics(y_true: np.ndarray, y_pred: np.ndarray) -> MetricsType:
    """Calculate various regression metrics.

    These are:
     * Mean Pinball Loss
     * SMAPE
     * Explained Variance
     * Max Error
     * MSE

    :param y_true: ground truth (correct) target values.
    :param y_pred: estimated target values.
    :return: a dictionary with the names of the metrics mapped to their values.
    """
    metrics = {
        "mean pinball loss": np.nan
        if any(np.isnan(y_pred))
        else mean_pinball_loss(y_true, y_pred),
        "SMAPE": smape(y_true, y_pred),
        "Explained Variance": np.nan
        if any(np.isnan(y_pred))
        else explained_variance_score(y_true, y_pred),
        "Max Error": np.nan if any(np.isnan(y_pred)) else max_error(y_true, y_pred),
        "MSE": np.nan if any(np.isnan(y_pred)) else mean_squared_error(y_true, y_pred),
    }

    return metrics


def report_metrics(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    pair_name: str,
    model_name: Optional[str] = None,
) -> str:
    """Calculate and report various regression metrics.

    These are:
        * Mean Pinball Loss
        * SMAPE
        * Explained Variance
        * Max Error
        * MSE

    :param y_true: ground truth (correct) target values.
    :param y_pred: estimated target values.
    :param pair_name: the name of the pool for which the metrics are reported.
    :param model_name: the name of the model for which the metrics are reported.
        The model's name will appear in the report's title.
    :return: a report string.
    """
    metrics = calc_metrics(y_true, y_pred)

    title = f"Pool {pair_name} metrics report"
    title += ":" if model_name is None else f" for {model_name}:"

    separator = "-" * len(list(title))

    report = f"\n{separator}\n{title}\n{separator}\n"

    for name, value in metrics.items():
        report += f"\t{name}: {value}"

    return report


def walk_forward_test(
    forecaster: Pipeline, y_test: np.ndarray, steps_forward: int = 1
) -> np.ndarray:
    """Test the given forecasting model, using the Direct Multi-step Forecast Strategy.

    :param forecaster: a `pmdarima` pipeline model.
    :param y_test: the test timeseries.
    :param steps_forward: how many timesteps the model will be predicting in the future.
    :return: a `numpy` array with the forecaster's predictions.
    """
    if steps_forward < 1:
        raise ValueError(
            f"Timesteps to predict in the future cannot be {steps_forward} < 1."
        )

    if steps_forward > len(y_test):
        warnings.warn(
            "Timesteps to predict in the future are larger than the number of test samples "
            f"while using the Direct Multi-step Forecast Strategy: {steps_forward} > {len(y_test)}"
        )

    y_pred = []
    for i in range(0, len(y_test), steps_forward):
        y_hat = predict_safely(forecaster, steps_forward)

        if steps_forward == 1:
            y_pred.append(y_hat)
        else:
            y_pred.extend(y_hat)

        forecaster.update(y_test[i : i + steps_forward])

    return np.asarray(y_pred)


def test_forecaster_per_pool(
    forecasters: PoolIdToForecasterType,
    y_train: PoolIdToTrainDataType,
    y_test: PoolIdToTrainDataType,
    steps_forward: int = 1,
) -> PoolIdToTestReportType:
    """Test the trained forecasters of the given pools and compare them with a Naive Baseline method."""
    return {
        id_: test_forecaster(
            forecaster,
            y_train[id_.replace(".joblib", ".csv")],
            y_test[id_.replace(".joblib", ".csv")],
            id_.replace(".joblib", ""),
            steps_forward,
        )
        for id_, forecaster in forecasters.items()
    }


def test_forecaster(
    forecaster: Pipeline,
    y_train: np.ndarray,
    y_test: np.ndarray,
    pair_name: str,
    steps_forward: int = 1,
) -> TestReportType:
    """Test the trained forecaster and compare it with a Naive Baseline method.

    :param forecaster: a `pmdarima` pipeline model.
    :param y_train: the train timeseries.
    :param y_test: the test timeseries.
    :param pair_name: the pair's name.
    :param steps_forward: how many timesteps the model will be predicting in the future.
    :return: a test report for each tested method/model.
    """
    # Get the current timestep, i.e., the last value of the training set.
    t0 = y_train[-1]

    # Get baseline's and model's predictions.
    report: Dict[str, Union[str, np.ndarray]] = {
        "Baseline": baseline(t0, y_test),
        "ARIMA": walk_forward_test(forecaster, y_test, steps_forward),
    }

    # Report baseline's and model's metrics.
    report.update(
        (
            reporting_model,
            report_metrics(
                y_test, cast(np.ndarray, model_predictions), pair_name, reporting_model
            ),
        )
        for reporting_model, model_predictions in report.items()
    )

    return cast(TestReportType, report)


def update_forecaster_per_pool(
    y: pd.DataFrame,
    forecasters: PoolIdToForecasterType,
) -> None:
    """Update the given forecasters.

    :param y: the data to use for the update.
    :param forecasters: the forecasters to update.
    """
    for id_ in forecasters.keys():
        apy = y.loc[y["id"] == id_.replace(".joblib", ".csv"), "APY"]
        forecasters[id_].update(apy)


def estimate_apy_per_pool(
    forecasters: PoolIdToForecasterType,
    steps_forward: int,
) -> pd.DataFrame:
    """Estimate the APY per pool.

    :param forecasters: the forecasters to update.
    :param steps_forward: the number of periods in the future to forecast.
    :return: a dataframe with the APY predictions per pool, for each step forward.
    """
    estimates = {}
    for id_, forecaster in forecasters.items():
        id_ = id_.replace(".joblib", "")
        estimates[id_] = predict_safely(forecaster, steps_forward)
    return pd.DataFrame(
        estimates, index=[f"Step{i + 1} into the future" for i in range(steps_forward)]
    )


def predict_safely(forecaster: Pipeline, steps_forward: int) -> Any:
    """
    Overcomes an issue of the `pmdarima` library.

    This has to be done because of a `pmdarima`'s issue: https://github.com/alkaline-ml/pmdarima/issues/404
    The issue is caused because of a check of the confidence interval after predicting.
    If the upper-lower bounds are `None`, then an error is raised from an assertion method.

    :param forecaster: a `pmdarima` pipeline model.
    :param steps_forward: how many timesteps the model will be predicting in the future.
    :return: the predicted values.
    """
    try:
        y_hat = forecaster.predict(steps_forward)
    except ValueError:
        y_hat = [np.nan] * steps_forward if steps_forward > 1 else np.nan

    return y_hat



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: acn.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\tacn.proto\x12\x12\x61\x65\x61.aea.acn.v1_0_0"\x92\x0b\n\nAcnMessage\x12P\n\x0c\x61\x65\x61_envelope\x18\x05 \x01(\x0b\x32\x38.aea.aea.acn.v1_0_0.AcnMessage.Aea_Envelope_PerformativeH\x00\x12T\n\x0elookup_request\x18\x06 \x01(\x0b\x32:.aea.aea.acn.v1_0_0.AcnMessage.Lookup_Request_PerformativeH\x00\x12V\n\x0flookup_response\x18\x07 \x01(\x0b\x32;.aea.aea.acn.v1_0_0.AcnMessage.Lookup_Response_PerformativeH\x00\x12H\n\x08register\x18\x08 \x01(\x0b\x32\x34.aea.aea.acn.v1_0_0.AcnMessage.Register_PerformativeH\x00\x12\x44\n\x06status\x18\t \x01(\x0b\x32\x32.aea.aea.acn.v1_0_0.AcnMessage.Status_PerformativeH\x00\x1a\xac\x01\n\x0b\x41gentRecord\x12\x12\n\nservice_id\x18\x01 \x01(\t\x12\x11\n\tledger_id\x18\x02 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\t\x12\x12\n\npublic_key\x18\x04 \x01(\t\x12\x17\n\x0fpeer_public_key\x18\x05 \x01(\t\x12\x11\n\tsignature\x18\x06 \x01(\t\x12\x12\n\nnot_before\x18\x07 \x01(\t\x12\x11\n\tnot_after\x18\x08 \x01(\t\x1a\x92\x03\n\nStatusBody\x12\x46\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x38.aea.aea.acn.v1_0_0.AcnMessage.StatusBody.StatusCodeEnum\x12\x0c\n\x04msgs\x18\x02 \x03(\t"\xad\x02\n\x0eStatusCodeEnum\x12\x0b\n\x07SUCCESS\x10\x00\x12\x1d\n\x19\x45RROR_UNSUPPORTED_VERSION\x10\x01\x12\x1c\n\x18\x45RROR_UNEXPECTED_PAYLOAD\x10\x02\x12\x11\n\rERROR_GENERIC\x10\x03\x12\x10\n\x0c\x45RROR_DECODE\x10\x04\x12\x1d\n\x19\x45RROR_WRONG_AGENT_ADDRESS\x10\n\x12\x1a\n\x16\x45RROR_WRONG_PUBLIC_KEY\x10\x0b\x12\x17\n\x13\x45RROR_INVALID_PROOF\x10\x0c\x12\x1c\n\x18\x45RROR_UNSUPPORTED_LEDGER\x10\r\x12\x1f\n\x1b\x45RROR_UNKNOWN_AGENT_ADDRESS\x10\x14\x12\x19\n\x15\x45RROR_AGENT_NOT_READY\x10\x15\x1aS\n\x15Register_Performative\x12:\n\x06record\x18\x01 \x01(\x0b\x32*.aea.aea.acn.v1_0_0.AcnMessage.AgentRecord\x1a\x34\n\x1bLookup_Request_Performative\x12\x15\n\ragent_address\x18\x01 \x01(\t\x1aZ\n\x1cLookup_Response_Performative\x12:\n\x06record\x18\x01 \x01(\x0b\x32*.aea.aea.acn.v1_0_0.AcnMessage.AgentRecord\x1ai\n\x19\x41\x65\x61_Envelope_Performative\x12\x10\n\x08\x65nvelope\x18\x01 \x01(\x0c\x12:\n\x06record\x18\x02 \x01(\x0b\x32*.aea.aea.acn.v1_0_0.AcnMessage.AgentRecord\x1aN\n\x13Status_Performative\x12\x37\n\x04\x62ody\x18\x01 \x01(\x0b\x32).aea.aea.acn.v1_0_0.AcnMessage.StatusBodyB\x0e\n\x0cperformativeb\x06proto3'
)


_ACNMESSAGE = DESCRIPTOR.message_types_by_name["AcnMessage"]
_ACNMESSAGE_AGENTRECORD = _ACNMESSAGE.nested_types_by_name["AgentRecord"]
_ACNMESSAGE_STATUSBODY = _ACNMESSAGE.nested_types_by_name["StatusBody"]
_ACNMESSAGE_REGISTER_PERFORMATIVE = _ACNMESSAGE.nested_types_by_name[
    "Register_Performative"
]
_ACNMESSAGE_LOOKUP_REQUEST_PERFORMATIVE = _ACNMESSAGE.nested_types_by_name[
    "Lookup_Request_Performative"
]
_ACNMESSAGE_LOOKUP_RESPONSE_PERFORMATIVE = _ACNMESSAGE.nested_types_by_name[
    "Lookup_Response_Performative"
]
_ACNMESSAGE_AEA_ENVELOPE_PERFORMATIVE = _ACNMESSAGE.nested_types_by_name[
    "Aea_Envelope_Performative"
]
_ACNMESSAGE_STATUS_PERFORMATIVE = _ACNMESSAGE.nested_types_by_name[
    "Status_Performative"
]
_ACNMESSAGE_STATUSBODY_STATUSCODEENUM = _ACNMESSAGE_STATUSBODY.enum_types_by_name[
    "StatusCodeEnum"
]
AcnMessage = _reflection.GeneratedProtocolMessageType(
    "AcnMessage",
    (_message.Message,),
    {
        "AgentRecord": _reflection.GeneratedProtocolMessageType(
            "AgentRecord",
            (_message.Message,),
            {
                "DESCRIPTOR": _ACNMESSAGE_AGENTRECORD,
                "__module__": "acn_pb2"
                # @@protoc_insertion_point(class_scope:aea.aea.acn.v1_0_0.AcnMessage.AgentRecord)
            },
        ),
        "StatusBody": _reflection.GeneratedProtocolMessageType(
            "StatusBody",
            (_message.Message,),
            {
                "DESCRIPTOR": _ACNMESSAGE_STATUSBODY,
                "__module__": "acn_pb2"
                # @@protoc_insertion_point(class_scope:aea.aea.acn.v1_0_0.AcnMessage.StatusBody)
            },
        ),
        "Register_Performative": _reflection.GeneratedProtocolMessageType(
            "Register_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _ACNMESSAGE_REGISTER_PERFORMATIVE,
                "__module__": "acn_pb2"
                # @@protoc_insertion_point(class_scope:aea.aea.acn.v1_0_0.AcnMessage.Register_Performative)
            },
        ),
        "Lookup_Request_Performative": _reflection.GeneratedProtocolMessageType(
            "Lookup_Request_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _ACNMESSAGE_LOOKUP_REQUEST_PERFORMATIVE,
                "__module__": "acn_pb2"
                # @@protoc_insertion_point(class_scope:aea.aea.acn.v1_0_0.AcnMessage.Lookup_Request_Performative)
            },
        ),
        "Lookup_Response_Performative": _reflection.GeneratedProtocolMessageType(
            "Lookup_Response_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _ACNMESSAGE_LOOKUP_RESPONSE_PERFORMATIVE,
                "__module__": "acn_pb2"
                # @@protoc_insertion_point(class_scope:aea.aea.acn.v1_0_0.AcnMessage.Lookup_Response_Performative)
            },
        ),
        "Aea_Envelope_Performative": _reflection.GeneratedProtocolMessageType(
            "Aea_Envelope_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _ACNMESSAGE_AEA_ENVELOPE_PERFORMATIVE,
                "__module__": "acn_pb2"
                # @@protoc_insertion_point(class_scope:aea.aea.acn.v1_0_0.AcnMessage.Aea_Envelope_Performative)
            },
        ),
        "Status_Performative": _reflection.GeneratedProtocolMessageType(
            "Status_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _ACNMESSAGE_STATUS_PERFORMATIVE,
                "__module__": "acn_pb2"
                # @@protoc_insertion_point(class_scope:aea.aea.acn.v1_0_0.AcnMessage.Status_Performative)
            },
        ),
        "DESCRIPTOR": _ACNMESSAGE,
        "__module__": "acn_pb2"
        # @@protoc_insertion_point(class_scope:aea.aea.acn.v1_0_0.AcnMessage)
    },
)
_sym_db.RegisterMessage(AcnMessage)
_sym_db.RegisterMessage(AcnMessage.AgentRecord)
_sym_db.RegisterMessage(AcnMessage.StatusBody)
_sym_db.RegisterMessage(AcnMessage.Register_Performative)
_sym_db.RegisterMessage(AcnMessage.Lookup_Request_Performative)
_sym_db.RegisterMessage(AcnMessage.Lookup_Response_Performative)
_sym_db.RegisterMessage(AcnMessage.Aea_Envelope_Performative)
_sym_db.RegisterMessage(AcnMessage.Status_Performative)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _ACNMESSAGE._serialized_start = 34
    _ACNMESSAGE._serialized_end = 1460
    _ACNMESSAGE_AGENTRECORD._serialized_start = 449
    _ACNMESSAGE_AGENTRECORD._serialized_end = 621
    _ACNMESSAGE_STATUSBODY._serialized_start = 624
    _ACNMESSAGE_STATUSBODY._serialized_end = 1026
    _ACNMESSAGE_STATUSBODY_STATUSCODEENUM._serialized_start = 725
    _ACNMESSAGE_STATUSBODY_STATUSCODEENUM._serialized_end = 1026
    _ACNMESSAGE_REGISTER_PERFORMATIVE._serialized_start = 1028
    _ACNMESSAGE_REGISTER_PERFORMATIVE._serialized_end = 1111
    _ACNMESSAGE_LOOKUP_REQUEST_PERFORMATIVE._serialized_start = 1113
    _ACNMESSAGE_LOOKUP_REQUEST_PERFORMATIVE._serialized_end = 1165
    _ACNMESSAGE_LOOKUP_RESPONSE_PERFORMATIVE._serialized_start = 1167
    _ACNMESSAGE_LOOKUP_RESPONSE_PERFORMATIVE._serialized_end = 1257
    _ACNMESSAGE_AEA_ENVELOPE_PERFORMATIVE._serialized_start = 1259
    _ACNMESSAGE_AEA_ENVELOPE_PERFORMATIVE._serialized_end = 1364
    _ACNMESSAGE_STATUS_PERFORMATIVE._serialized_start = 1366
    _ACNMESSAGE_STATUS_PERFORMATIVE._serialized_end = 1444
# @@protoc_insertion_point(module_scope)
